"""Bileto Models

This file defines the two DB schemas, Request and Comment, and various
metadata related to those.
"""

import re

from os import environ
from os.path import join
from datetime import datetime
from functools import lru_cache
from fcntl import LOCK_EX, LOCK_NB, flock

from sqlalchemy import BigInteger, desc, text
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import relationship
from sqlalchemy.dialects import sqlite
from flask import request, session

from bileto.app import db


# SQLite can't handle the BigInts, so we overload that for unit tests.
# Behaves like proper bigint should in production postgres.
BigInt = BigInteger().with_variant(sqlite.INTEGER(), 'sqlite')


CLEAR = {'Failed to build', 'Preparing packages', 'Currently building'}
HIDDEN = ('Landed', 'Abandoned')
WORD = re.compile(r'\W+')
COMMENT = re.compile(r'#.*(\n|$)')
COMMASPACE = ', '
SLASH = '/'
COMMA = ','
SPACE = ' '
EMPTY = ''
PLUS = '+'
N = '\n'

AUTOPKGTEST = """Automated Test Results

Automatically filled with links to britney excuses.html including \
autopkgtest results."""

BRITNEY_SIGNOFF = """Automated Signoff

Indicates if britney has approved the silo based on autopkgtests & other \
factors. Check the excuses page for details."""

CREATOR = """Request Creator

Launchpad nickname of the person who created this request."""

DESCRIPTION = """Landing Description

Briefly describe what bug you're fixing or feature you're adding."""

DISTRO = """Target Distribution

What distribution to build for, eg, ubuntu or ubuntu-rtm."""

JOB_LOG = """URL of full status log

This is the full log leading to the creation of the current status."""

LANDERS = """Lander IRC Nicks

Space separated list of IRC nicks of the people responsible for the silo."""

LANDER_SIGNOFF = """Lander Signoff

Indicates whether or not the silo has been tested by the upstream lander."""

MANUAL_SOURCES = """Source Package Names

Automatically-discovered list of source packages found in PPA."""

MANUAL_URLS = """Manual Download URLs

Place links to click packages and tarballs here."""

MERGES = """Merge Proposal URLs

Space or newline separated list of MERGE URLs. No branches!"""

PPA = """Destination

Whether silo targets ubuntu archive or overlay PPA. No effect on dual silos."""

QA = """QA Signoff

Indicates whether or not the silo has been approved by QA team."""

SERIES = """Target Series

What series to build for, eg, wily, vivid, or yakkety+xenial+vivid."""

SILONAME = """Name of Assigned Silo

Set automatically by Bileto, do not touch."""

STATUS = """Status

Set automatically by Bileto, do not touch."""

SYNC_SOURCE = """Sync Source

Where should this silo be synced from? Leave blank if not syncing.
Valid values are '[silo number]', '[distro],[series]', '[ppa],[series]'
or 'stable-overlay,[series]'."""

TEST_PLAN = """Test Plan

Link to a document that explains the steps to verify this silo."""

SUMMARY = """Packages:
{packages}

More info:
{REQUEST_URL_ROOT}{permalink}
"""

COLORS = dict(
    blue={'granted'},
    green={'landed'},
    grey={'abandoned'},
    olive={'rebuild', 'needs', 'not'},
    red={'caught', 'uncaught', 'exception', 'fail', 'failed', 'error',
         'unacceptable'},
)

BAD_DISTRO = '{} is not a supported distro.'
MIGRATE_TO_YAKKETY = """This ticket must be migrated to yakkety+xenial+vivid\
 before it can be built."""


def right_now(*ignore):
    """Return the current time in UTC."""
    return datetime.utcnow()


def splitter(raw):
    """Parse arbitrary spaces & commas into a list."""
    return raw.replace(COMMA, SPACE).split()


class BiletoError(Exception):
    """Raising this exception logs the error message without a traceback."""
    pass


class BiletoModel:
    """Base class for db models."""
    verboten = set()
    hidden = set()

    def __getitem__(self, key):
        """Expose object attributes like dict values."""
        return getattr(self, key)

    def keys(self):
        """Identify what db columns we have."""
        return sorted(set(inspect(self).attrs.keys()) - self.hidden)

    def update(self, **kwargs):
        """Update values of arbitrary db fields."""
        keys = self.keys()
        self.__init__(**{
            key: val for key, val in kwargs.items()
            if key in keys and key not in self.verboten
        })

    @classmethod
    @lru_cache()
    def docs(cls):
        """Generate a dictionary representing db field documentation."""
        return {key: val.doc for key, val in inspect(cls).attrs.items()}

    def commit(self):
        """Persist changes in db."""
        db.session.commit()


class Request(db.Model, BiletoModel):
    """Represents one landing ticket in the database."""
    request_id = db.Column(BigInt, primary_key=True, doc='Request ID')
    date = db.Column(db.DateTime, default=right_now, doc='Creation Date')

    test_plan = db.Column(db.Text, default=EMPTY, doc=TEST_PLAN)
    description = db.Column(db.Text, default=EMPTY, doc=DESCRIPTION)
    download_links = db.Column(db.Text, default=EMPTY, doc=MANUAL_URLS)
    merge_proposals = db.Column(db.Text, default=EMPTY, doc=MERGES)

    status_color = db.Column(db.Text, default=EMPTY, doc='Status Color')
    artifacts = db.Column(db.Text, default=EMPTY, doc='Artifacts')

    autopkgtest = db.Column(db.Text, default=EMPTY, doc=AUTOPKGTEST)
    job_log = db.Column(db.Text, default=EMPTY, doc=JOB_LOG)
    creator = db.Column(db.Text, default=EMPTY, doc=CREATOR)
    landers = db.Column(db.Text, default=EMPTY, doc=LANDERS)
    siloname = db.Column(db.Text, default=EMPTY, doc=SILONAME)
    status = db.Column(
        db.Text, nullable=False, default='New', doc=STATUS)
    sync_request = db.Column(
        db.Text, default=EMPTY, doc=SYNC_SOURCE)
    sources = db.Column(
        db.Text, default=EMPTY, doc=MANUAL_SOURCES)
    published_versions = db.Column(db.Text, default=EMPTY, doc='Deprecated.')

    distribution = db.Column(
        db.Text, default=EMPTY, doc=DISTRO)
    series = db.Column(db.Text, default=EMPTY, doc=SERIES)
    dest = db.Column(db.Text, default=EMPTY, doc=PPA)
    lander_signoff = db.Column(db.Text, default=EMPTY, doc=LANDER_SIGNOFF)
    britney_signoff = db.Column(db.Text, default=EMPTY, doc=BRITNEY_SIGNOFF)
    qa_signoff = db.Column(db.Text, default=EMPTY, doc=QA)

    search = db.Column(db.Text, default=EMPTY, doc='Searchable Field')

    comments = relationship(
        'Comment', order_by=desc(text('Comment.id')), doc='Audit Log')

    verboten = {'comments', 'date', 'request_id', 'search', 'status_color'}

    @property
    def sources_set(self):
        """Return source package names as a set."""
        return set(splitter(self.sources))

    @property
    def sources_sorted(self):
        """Return source package names as a sorted list."""
        return sorted(splitter(self.sources))

    @property
    def merge_proposals_list(self):
        """Return merge proposals as a list, ignoring comments."""
        return splitter(COMMENT.sub(N, self.merge_proposals))

    @property
    def path(self):
        """Return a permalink for linking to this ticket."""
        return '/#/ticket/{}'.format(self.request_id)

    @classmethod
    def new(cls):
        """Create a new Request with some values filled in."""
        user = session.get('nickname', EMPTY)
        return cls(
            distribution='ubuntu',
            series='yakkety+xenial+vivid',
            qa_signoff='Required',
            landers=session.get('ircnick', user))

    def validate(self):
        """Raise an exception if this ticket is misconfigured."""
        if self.sync_request:
            raise BiletoError('Syncs are no longer supported.')
        if not (self.merge_proposals or self.sources):
            raise BiletoError('Need either merges or sources.')
        if self.distribution not in ('ubuntu',):
            raise BiletoError(BAD_DISTRO.format(self.distribution))
        if self.series == 'xenial+vivid':
            raise BiletoError(MIGRATE_TO_YAKKETY)

    def summarize(self):
        """Return a text summary of this ticket."""
        return SUMMARY.format(
            packages=N.join(
                'https://launchpad.net/ubuntu/+source/' + source
                for source in self.sources_sorted),
            requestid=self.request_id,
            permalink=self.path,
            **environ  # Trailing comma is syntax error in py3.4
        )

    def publishable(self):
        """Identify if this request is appropriate to publish or not."""
        if self.status in HIDDEN:
            return False
        return not {
            self.lander_signoff, self.britney_signoff, self.qa_signoff
        } - {'Approved', 'N/A'}

    def compute_status_color(self, tokens):
        """Identify what color the status message should be."""
        if self.publishable():
            return 'blue'
        for signoff in ('lander', 'britney', 'qa'):
            tokens.add((getattr(self, signoff + '_signoff') or EMPTY).lower())
        for color, keywords in reversed(sorted(COLORS.items())):
            if tokens & keywords:
                return color
        return 'black'

    def clear_signoffs(self, tokens):
        """Clear signoffs in various circumstances."""
        status = self.status or EMPTY
        lander_fail = self.lander_signoff != 'Approved'
        if lander_fail or any(clear in status for clear in CLEAR):
            self.lander_signoff = EMPTY
            self.britney_signoff = EMPTY
            self.qa_signoff = 'Required'
            self.autopkgtest = EMPTY

    def compute_qa_signoff(self):
        """Enforce QA signoff based on various conditions."""
        current = self.qa_signoff
        if self.download_links:
            return current  # TODO: Figure out a better way to handle tarballs
        if current in ('Approved', 'Failed'):
            return current
        dest = self.dest or EMPTY
        series = self.series or EMPTY
        if 'stable-phone-overlay' in dest or 'vivid' in series:
            if self.lander_signoff == self.britney_signoff == 'Approved':
                return 'Ready'
            return 'Required'
        return 'N/A'

    def compute_britney_signoff(self):
        """Notify if this ticket is properly queued for britney."""
        current = self.britney_signoff
        if current == EMPTY and self.lander_signoff == 'Approved':
            if SLASH in self.siloname and 'Success' in self.status:
                return 'Queued'
        return current

    def __lt__(self, other):
        """Allow requests to be sorted."""
        return self.request_id < other.request_id

    def _audit(self, status, job_log, author='bileto-bot', **kwargs):
        """Create a Comment object which records a change in the ticket."""
        db.session.add(Comment(
            log_url=job_log,
            published_versions=kwargs.get('published_versions', EMPTY),
            siloname=kwargs.get('siloname', self.siloname),
            request_id=self.request_id,
            author=author,
            text=status,
        ))

    def audit(self, **kwargs):
        """Create audit log messages as appropriate."""
        no_audit = self.verboten | {'author', 'job_log'}
        nick = environ.get('BLAME') or session.get('nickname')
        kwargs['author'] = nick or kwargs.get('author') or 'bileto-bot'
        status = kwargs.get('status')
        if status and status != self.status:
            self._audit(**kwargs)
        else:
            comment = dict(kwargs)
            comment.setdefault('job_log', self.path)
            for key, val in comment.items():
                if key not in no_audit and hasattr(self, key):
                    if self[key] != val:
                        explicit = '{}: {}'.format(key, val)
                        brief = 'Updated {}.'.format(key)
                        self._audit(
                            status=explicit if len(explicit) < 100 else brief,
                            **comment)

    def update(self, **kwargs):
        """Ensure changes are logged & compute computed values."""
        self.audit(**kwargs)
        super().update(**kwargs)
        tokens = set(WORD.split((self.status or EMPTY).lower()))
        keys = sorted(set(super().keys()) - self.verboten)
        self.search = N.join([str(self[key] or EMPTY) for key in keys])
        self.clear_signoffs(tokens)
        self.qa_signoff = self.compute_qa_signoff()
        self.britney_signoff = self.compute_britney_signoff()
        self.status_color = self.compute_status_color(tokens)
        if self.status in HIDDEN:
            # These links will just 404 anyway.
            self.artifacts = EMPTY
            self.autopkgtest = EMPTY

    def set_status(self, message):
        """Update status immediately."""
        self.update(
            status=message,
            job_log=environ.get('LOG_PATH') or
            environ.get('REQUEST_PATH') or
            request.path)
        self.commit()
        return message

    @property
    @lru_cache()
    def lockfile(self):
        """Identify what file we lock on."""
        return join('/tmp', 'bileto-ticket-{}'.format(self.request_id))

    @property
    @lru_cache()
    def lock_fd(self):
        """Return the file descriptor that we lock on."""
        print(self.lockfile)
        return open(self.lockfile, 'w')

    def lock(self):
        """Prevent multiple jobs from running simultaneously."""
        print(self.lock_fd)
        flock(self.lock_fd, LOCK_EX | LOCK_NB)

    # TODO: not needed yet
    # def release(self):
    #     """Release the lock."""
    #     self.lock_fd.close()
    #     self.__class__.lock_fd.fget.cache_clear()


class Comment(db.Model, BiletoModel):
    """Stores one comment on a landing request."""
    id = db.Column(BigInt, primary_key=True)
    request_id = db.Column(
        BigInt, db.ForeignKey(Request.request_id), nullable=False)
    author = db.Column(db.Text, nullable=False)
    text = db.Column(db.Text, default=EMPTY)
    date = db.Column(db.DateTime, default=right_now)

    log_url = db.Column(db.Text, default=EMPTY)
    published_versions = db.Column(db.Text, default=EMPTY)
    siloname = db.Column(db.Text, default=EMPTY)

    verboten = {'id', 'date'}
