"""Bileto Streams

Run a subprocess and stream the output to an HTTP client.
"""

import re

from os.path import basename, join
from os import environ, makedirs, walk

from glob import glob
from time import sleep
from subprocess import Popen
from contextlib import suppress
from fcntl import flock, LOCK_SH, LOCK_NB

from flask import (
    abort,
    redirect,
    render_template,
    request,
    session,
)

from bileto.app import app
from bileto.models import Request
from bileto.actions import Actions
from bileto.settings import Config
from bileto.auth import ADMIN_TEAMS, require_team


UNDERSCORE = '_'
bEMPTY = b''
EMPTY = ''
SPACE = ' '
SLASH = '/'
COMMA = ','
STAR = '*'
DOT = '.'

# Only ADMIN_TEAMS can finalize, all others can be done by anybody.
TEAMS = dict(finalize=ADMIN_TEAMS)

# TODO: python3-bleach does this better, but isn't in trusty. Migrate to xenial
# then drop this ugly regex.
URL = re.compile(
    r'(?<!")(https?://|/(#|log)/)[-A-Z0-9+&@#/%?=~_|!:,.;]*[A-Z0-9+&@#/%=~_|]',
    flags=re.IGNORECASE)
REGEX_PROBLEMS = ('&gt', '&#39')


def shorter(matchobj, suffix=EMPTY):
    """HTMLify a matched URL."""
    url = matchobj.group(0)
    for prob in REGEX_PROBLEMS:
        if url.endswith(prob):
            url = url[:-len(prob)]
            suffix += prob
    parts = url.split(SLASH)[5 if 'AUTH_' in url else 3:]
    short = SPACE.join(parts).replace(UNDERSCORE, SPACE)
    return '<a href="{}">{}</a>{}'.format(url, short, suffix)


def linkify(raw):
    """Convert URLs to HTML links in strings."""
    return URL.sub(shorter, raw)


def logview(requestid, action, num, level, logdata, running, **kwargs):
    """Render the template used for viewing logs."""
    return render_template(
        'logview.html',
        toggle=Config.LEVELS[level == Config.LEVELS[0]],
        title='{} {} of ticket {}'.format(
            action.capitalize(), num, requestid),
        **locals())


def assert_action(action):
    """Abort if action is invalid."""
    if action not in ('build', 'diff'):
        abort(404)


def list_packages(action, requestid):
    """Return a list of packages to prompt the user with."""
    if action == 'build':
        ticket = Request.query.get(requestid)
        if ticket:
            sources = ticket.sources_sorted
            # Hide GLES because those are created automatically from non-gles.
            return [s for s in sources if not s.endswith(Config.GLES)]
    return []


def log_meta(files):
    """Return a dict representing log metadata."""
    data = {}
    for name in files:
        key, _, val = basename(name).partition(UNDERSCORE)
        data[key] = val
    return data


def log_scan(path):
    """Read logs & metadata from disk."""
    makedirs(path, exist_ok=True)
    data = {}
    for root, dirs, files in walk(path):
        if len(files) >= 3:  # info, debug, stamp
            lognum = int(basename(root))
            data[lognum] = log_meta(files)
    return reversed(sorted(data.items()))


def is_running(lock):
    """Determine if a given log file is actively being written to."""
    with suppress(BlockingIOError):
        flock(lock, LOCK_SH | LOCK_NB)
        return False
    return True


@app.route('/log/', methods=['GET'])
@app.route('/log/<int:requestid>/', methods=['GET'])
def list_all_logs(requestid=None):
    """List all logs in chronological order."""
    paths = glob(join(Config.HOME, 'log', STAR, STAR, STAR, 'info'))
    if requestid:
        needed = '/log/{}/'.format(requestid)
        paths = [path for path in paths if needed in path]
    paths = [path for path in paths if 'latest' not in path]
    meta = {path: log_meta(glob(path.replace('info', STAR))) for path in paths}
    logs = {path: path.replace(Config.HOME, EMPTY) for path in paths}
    paths = sorted(paths, key=lambda p: meta[p].get('stamp'), reverse=True)
    return render_template('loglist.html', title='All Jobs', **locals())


@app.route('/log/<int:requestid>/<action>/', methods=['GET'])
def configure_action(requestid, action):
    """List past logs & the form for triggering new actions."""
    assert_action(action)
    logs = log_scan(join(Config.HOME, request.path.strip(SLASH)))
    return linkify(render_template(
        'trigger.html',
        title='{} ticket {}'.format(action.capitalize(), requestid),
        post=join(request.path, 'latest/'),
        packages=list_packages(action, requestid),
        message=getattr(Actions, action).__doc__,
        **locals()))


@app.route('/log/<int:requestid>/<action>/latest/', methods=['POST'])
def perform_action(requestid, action, level=Config.LEVELS[0]):
    """Call subprocess and stream output directly to clients."""
    assert_action(action)
    require_team(TEAMS.get(action))
    env = environ.copy()
    env.update(
        ACTION=action,
        BLAME=session['nickname'],
        BLAME_EMAIL=session['email'],
        BLAME_FULLNAME=session['fullname'],
        HOME=Config.HOME,  # FIXME: not sure why this isn't set for us...
        INCLUDES=SPACE.join(request.form.getlist('includes')),
        PYTHONPATH=Config.ROOT,
        REQUEST_PATH=request.path,
        REQUEST_URL_ROOT=request.url_root.strip(SLASH),
        REQUEST_URL=request.url,
        REQUESTID=str(requestid),
        TEAMS=COMMA.join(session.get('teams', [])),
    )
    path = join(Config.ROOT, 'bileto', 'worker', 'main.py')
    Popen((path,), env=env)
    sleep(5)
    return redirect(request.path, code=303)


@app.route('/log/<int:requestid>/<action>/<num>/', methods=['GET'])
@app.route('/log/<int:requestid>/<action>/<num>/<level>/', methods=['GET'])
def show_log(requestid, action, num, level=Config.LEVELS[0]):
    """Show one full log."""
    assert_action(action)
    if level not in Config.LEVELS:
        abort(400)
    path = join(Config.HOME, 'log', str(requestid), action, num, level)
    with suppress(FileNotFoundError):
        with open(path, encoding='utf-8') as data:
            running = is_running(data)
            return linkify(logview(logdata=data.read(), **locals()))
    abort(404)
