"""Bileto Log

This module defines the following logging behavior:

1. Log INFO messages to a file
2. Log INFO+DEBUG messages to a different file

This way, you can always access DEBUG messages even retroactively, and there is
no longer any need for having a 'DEBUG' flag in our scripts.

This module should only be used by the subprocess spawned by streams.py, as
the parent process won't have the necessary environment variables set:

Environment variables:
    REQUESTID
        The id of the ticket we are acting on, an integer.

    ACTION
        One of 'build', 'publish', 'finalize', or 'abandon'.
"""

import logging

from os.path import join
from os import environ, listdir, makedirs, rename, symlink

from fcntl import flock, LOCK_EX
from functools import lru_cache
from datetime import datetime
from pprint import pformat
from time import gmtime

from bileto.settings import Config

FORMATTER = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
logging.Formatter.converter = gmtime

EMPTY = ''


def make_handler(logger, data, level):
    """Create & add a handler to a logger."""
    handler = logging.StreamHandler(data)
    handler.setLevel(level)
    handler.setFormatter(FORMATTER)
    logger.addHandler(handler)


class log:
    """Lazy load logging config for easier testing."""
    _dir = join(Config.HOME, 'log')
    _action = environ.get('ACTION')
    _requestid = environ.get('REQUESTID')

    def _find_path(self, *args):
        """Join helper."""
        assert all(args) and self._requestid and self._action
        return join(self._dir, self._requestid, self._action, *args)

    def _find_dir(self, *args):
        """Create directories."""
        path = self._find_path(*args)
        makedirs(path, exist_ok=True)
        return path

    @property
    @lru_cache()
    def logdir(self):
        """What directory are we logging to?"""
        present = listdir(self._find_dir())
        latest = [int(name) for name in present if name.isdigit()] or [0]
        logdir = self._find_dir(str(max(latest) + 1))
        environ['LOG_PATH'] = logdir.replace(Config.HOME, EMPTY)
        return logdir

    @property
    @lru_cache()
    def _logger(self):
        """Set up loggers for INFO and DEBUG files, and STDOUT."""
        logger = logging.getLogger('bileto')
        logger.setLevel(logging.DEBUG)
        self.link('latest')
        self.touch(datetime.utcnow().strftime('stamp_%Y-%m-%d %H:%M:%S'))
        for level in Config.LEVELS:
            logname = join(self.logdir, level)
            data = open(logname, 'x', encoding='utf-8')
            flock(data, LOCK_EX)
            make_handler(logger, data, getattr(logging, level.upper()))
        return logger

    def __getattr__(self, attr):
        """Proxy all calls through to logger object."""
        return getattr(self._logger, attr)

    def vars(self, value, message=''):
        """Debug local variables."""
        self.debug('%s\n%s', message, pformat(value))

    def finalize(self, success, time):
        """Record whether the job was a success or not."""
        self.touch('success_{}'.format(success))
        self.touch('elapsed_{}'.format(time))
        self.link('latestSuccess' if success else 'latestFailure')

    def touch(self, filename):
        """Create an empty file or update mtime of file."""
        open(join(self.logdir, filename), 'a').close()

    def link(self, name):
        """Create a symlink."""
        dest = self._find_path(name)
        tmp = dest + '.tmp'
        symlink(self.logdir, tmp)
        rename(tmp, dest)

log = log()  # Single global instance
