#!/usr/bin/python3 -u
# IMPORTANT: Output must be unbuffered ('-u') for streaming purposes!

"""Bileto Worker Script

This script dispatches the Build, Publish, Finalize, and Abandon actions that
can be performed on a Bileto ticket.

This script is invoked as a subprocess by bileto/streams.py in order to do
work smoothly in parallel with flask's streaming abilities.

Environment variables:
    REQUESTID
        The id of the ticket we are acting on, an integer.

    ACTION
        One of 'build', 'publish', 'finalize', or 'abandon'.
"""

import sys

from shutil import rmtree
from tempfile import mkdtemp
from datetime import datetime
from contextlib import suppress
from urllib.parse import urlencode
from os import chdir, environ, getcwd
from lazr.restfulclient import resource

from bileto.models import BiletoError, Request
from bileto.streams import Actions

from bileto.worker.log import log
from bileto.worker.manager import Manager


# FIXME: This is fixed in xenial but still needed for trusty.
# https://bugs.launchpad.net/launchpadlib/+bug/1465279
# https://bugs.launchpad.net/lazr.restfulclient/+bug/1425609
with suppress(AttributeError):
    resource.urllib.urlencode = urlencode

PLUS = '+'


def main():
    """Initiate action and handle errors."""
    start = datetime.utcnow()
    success = False
    ticket = None
    chdir(mkdtemp('-bileto-workspace'))
    try:
        log.info('Job started by %s.', environ['BLAME'])
        environ.update(
            DEBEMAIL=environ['BLAME_EMAIL'],
            DEBFULLNAME=environ['BLAME_FULLNAME'],
        )
        log.vars(dict(environ), 'Env')
        requestid = environ['REQUESTID']
        ticket = Request.query.get(requestid)
        if ticket:
            ticket.lock()
            environ.update(PRIMARY=ticket.series.partition(PLUS)[0])
            action = environ['ACTION']
            ticket.set_status(Actions.statuses[action])
            getattr(Actions, action)(Manager(ticket))
            success = True
            return 0
        log.error('No ticket found with id %s!', requestid)
        return 1
    except BiletoError as err:
        log.error(err)
        if ticket:
            ticket.set_status(str(err))
        return 2
    except BlockingIOError:
        log.error('This ticket is busy. Try again later.')
        return 3
    except Exception:
        log.exception('Unhandled exception')
        return 4
    finally:
        time = datetime.utcnow() - start
        log.info('{} in {}'.format('Succeeded' if success else 'Failed', time))
        log.finalize(success, time)
        rmtree(getcwd(), ignore_errors=True)


if __name__ == '__main__':
    sys.exit(main())  # pragma: no cover
