"""Bileto Worker Manager

Manages the instantiation of Package classes and passing messages to them.

Environment Variables:
    INCLUDES
        Comma- or space-separated list of what packages to consider.

    REQUEST_URL_ROOT
        Where to anchor permalinks to, including protocol, ip/domain, and port.
"""

from os import environ
from contextlib import suppress
from functools import lru_cache
from collections import defaultdict

from lazr.restfulclient.errors import BadRequest, NotFound

from bileto.lplib import lp
from bileto.settings import Config
from bileto.models import BiletoError, splitter

from bileto.worker.log import log
from bileto.worker.gles import GLES
from bileto.worker.merge import Merge
from bileto.worker.manual import Manual
from bileto.worker.vcs import parallelize
from bileto.worker.package import Package
from bileto.worker.secondary import Secondary

MERGE_ORDER = """Merges have been sorted by prerequisite \
and will be merged in this order:
%s
"""

SLASH = '/'
SPACE = ' '
EMPTY = ''
PLUS = '+'
N = '\n'


def NOP(*args, **kwargs):
    """Do nothing."""


def name_ppa(requestid):
    """Convert the requestid into a legal lp ppa name."""
    return str(requestid).zfill(2)


def pretty_instances(items):
    """Pretty print instantiated instances."""
    by_dest = defaultdict(list)
    for name, obj in sorted(items):
        by_dest[obj.dest.web_link].append(name)
    for dest, names in sorted(by_dest.items()):
        log.info('Packages targetting %s:\n%s\n', dest, N.join(names))


class Manager:
    """Manage the instantiation of Package objects."""
    def __init__(self, ticket):
        self.ticket = ticket

    @lru_cache()
    def merges(self):
        """Instantiate all lp merge objects."""
        merges = Merge.prepare_all(self.ticket.merge_proposals_list)
        if merges:
            log.info(MERGE_ORDER, Merge.stringify(merges))
        return merges

    @lru_cache()
    def ppa_sources(self):
        """Instantiate all packages in PPA."""
        return self.ppa.getPublishedSources()

    @lru_cache()
    def ppa_names(self):
        """Identify names of all PPA packages."""
        return {source.source_package_name
                for source in self.ppa_sources()
                if source.status in ('Pending', 'Published')}

    @lru_cache()
    def names(self):
        """Choose which package names to act upon."""
        names = environ['INCLUDES']
        if names:
            return set(splitter(names))
        srcs = self.ticket.sources_set | set(self.merges()) | self.ppa_names()
        self.ticket.sources = SPACE.join(sorted(srcs))
        return srcs

    @lru_cache()
    def types(self):
        """Choose which classes to use."""
        # Only report non-gles sources as -gles twins are added automatically
        all_sources = {n for n in self.names() if not n.endswith(Config.GLES)}
        merges = set(self.merges())
        return {
            Merge: all_sources & merges,
            Manual: all_sources - merges,
        }

    @lru_cache()
    def packages(self):
        """Instantiate Package objects."""
        distro = self.ticket.distribution
        serieses = self.serieses
        ppa = self.ppa
        for kind, names in self.types().items():
            gles_kind = GLES if kind is Merge else Manual
            second_kind = Secondary if kind is Merge else Manual
            dest = self.dest
            for series in serieses:
                for name in names:
                    kind(name, distro, series, dest, ppa)
                    if name in GLES.TWINS:
                        gles_name = name + Config.GLES
                        gles_kind(gles_name, distro, series, dest, ppa)
                kind = second_kind
                dest = self.overlay
        if not Package.instances:
            raise BiletoError('No packages are being considered!')
        pretty_instances(Package.instances.items())
        return Package.instances

    def __getattr__(self, attr):
        """Pass method calls down to all instances."""
        def passthrough(*args, **kwargs):
            """Pass arguments down to methods."""
            log.debug('Beginning step: {}'.format(attr.upper()))
            getattr(Package, 'pre_' + attr, NOP)(self.ticket)
            parallelize([getattr(package, attr)(*args, **kwargs)
                         for name, package in self.packages().items()])
            getattr(Package, 'post_' + attr, NOP)(self.ticket)
            return self
        return passthrough

    def diff(self):
        """Force diffing of all packages."""
        environ.update(INCLUDES=self.ticket.sources)
        for func in (Manager.names, Manager.types, Manager.packages):
            func.cache_clear()
        self.__getattr__('diff')()

    def nuke(self):
        """Delete PPA."""
        self.delete_ppa(self.ticket.request_id)
        # TODO: Delete swift container

    @property
    @lru_cache()
    def ppa_ephemeral(self):
        """Create PPA, but only if necessary. TODO: Not used yet."""
        number = self.ticket.request_id
        distro = self.ticket.distribution
        ppa = self.get_ppa(number)
        if ppa:
            self.ticket.siloname = '{}/{}'.format(distro, ppa.name)
            return ppa
        return self.make_ppa(number, self.ticket.distribution, '{}{}'.format(
            environ['REQUEST_URL_ROOT'], self.ticket.path))

    @property
    @lru_cache()
    def ppa(self):
        """Use the PPA assigned by Jenkins. TODO: switch to ephemeral."""
        name = self.ticket.siloname
        if name:
            return self.get_ppa(name.rpartition(SLASH)[-1])
        raise BiletoError('You must assign this ticket before building it')

    @property
    @lru_cache()
    def dest(self):
        """Get object representing destination archive."""
        if self.ticket.series == 'vivid':
            return self.overlay
        if PLUS in self.ticket.series:
            return lp.ubuntu.main_archive
        if self.ticket.dest:
            team, distro, name = self.ticket.dest.split(SLASH)
            return lp.load('~{}/+archive/{}/{}'.format(team, distro, name))
        return lp.ubuntu.main_archive

    @property
    @lru_cache()
    def overlay(self):
        """Get object representing overlay ppa."""
        return lp.load('~{}/+archive/{}/stable-phone-overlay'.format(
            Config.ppa_team, self.ticket.distribution))

    @lru_cache()
    def get_series(self, name):
        """Get object representing ubuntu series."""
        return lp.load('{}/{}'.format(self.ticket.distribution, name))

    @property
    @lru_cache()
    def serieses(self):
        """Get objects representing ubuntu series in this ticket."""
        return [self.get_series(s) for s in self.ticket.series.split(PLUS)]

    def get_ppa(self, name):
        """Return the PPA object."""
        with suppress(NotFound, AttributeError):
            ppa = lp.ppa_team.getPPAByName(name=name_ppa(name))
            log.info('PPA: %s/+packages.', ppa.web_link)
            return ppa

    def make_ppa(self, number, distro, description):
        """Unconditionally create an ephemeral PPA."""
        log.info('Making PPA %s.', number)
        lp.ppa_team.createPPA(
            description=description,
            distribution=lp.distributions[distro],
            displayname='Bileto {}'.format(number),
            name=name_ppa(number),
        )
        # TODO: Needs to be devirted and all arches enabled.
        return self.get_ppa(number)

    def delete_ppa(self, number):
        """Unconditionally delete an ephemeral PPA."""
        log.info('Deleting PPA %s.', number)
        try:
            self.get_ppa(name_ppa(number)).lp_delete()
        except BadRequest as err:
            # Will log 'Archive already deleted.' if deleted but not gone yet
            log.warning(err.content.decode('utf-8'))
        except AttributeError:
            pass  # already gone for real
