"""Bileto Merges

Everything to do with handling of launchpad MPs.
"""

from os import environ
from asyncio import coroutine
from contextlib import suppress
from collections import defaultdict
from functools import lru_cache

from lazr.restfulclient.errors import NotFound

from bileto.lplib import lp
from bileto.models import BiletoError

from bileto.worker.vcs import VCS, VCSError, parallelize
from bileto.worker.package import Package
from bileto.worker.version import V
from bileto.worker.log import log

NOT_A_MERGE = 'Failed to understand "{}". Is it a merge?'
COMMASPACE = ', '
EMPTY = ''
LT = '<'
NN = '\n\n'
N = '\n'


class SortedMPList(list):
    """A list that sorts MPs by prerequisite branches while appending."""

    def __init__(self):
        super().__init__()
        self.prereqs = {}

    def append(self, item):
        """Append item in the correct order."""
        self.prereqs[item.prerequisite_branch_link] = item
        if self:
            source = item.source_branch_link
            if source in self.prereqs:
                index = self.index(self.prereqs[source])
                self.insert(index, item)
                log.vars(locals(), 'Inserting sortedly!')
                return
        super().append(item)


@coroutine
def get_source(result, targets, trunk, merges):
    """Arrange trunk name and merges list by source package name."""
    source = yield from VCS.source_name(trunk)
    targets[source].append(trunk)
    result[source] = merges


def parse_authors(rawthors):
    """Identify names & emails from a list of authors."""
    authors = set(rawthors.split(COMMASPACE))
    names, emails = [], []
    for author in sorted(authors):
        name, _, email = author.partition(' <')
        name = name.strip()
        email = email.strip(' <>')
        if name:
            names.append(name)
        if email:
            emails.append(email)
    log.vars(locals(), 'Parsed authors')
    return COMMASPACE.join(names), COMMASPACE.join(emails)


def format_bugs(bugs):
    """Pretty-print bug list."""
    ids = {'#{}'.format(bug.id) for bug in bugs}
    return ' (LP: {})'.format(', '.join(sorted(ids))) if ids else ''


class Merge(Package):
    """Define the steps unique to packages built from merge proposals."""
    all_mps = {}
    target = None

    @classmethod
    def prepare_all(cls, urls):
        """Process MPs for silo config."""
        log.info('Determining source package names from MPs...')
        by_trunk = defaultdict(SortedMPList)
        for url in urls:
            try:
                assert url.startswith(lp.MP_ROOT)
                merge = lp.load(url)
                by_trunk[merge.target_branch.display_name].append(merge)
            except (AssertionError, AttributeError, NotFound) as err:
                log.error(err)
                raise BiletoError(NOT_A_MERGE.format(url))
        log.vars(by_trunk, 'Merges arranged by trunk')
        cls.all_mps = {}
        targets = defaultdict(list)
        parallelize([get_source(cls.all_mps, targets, trunk, merges)
                     for trunk, merges in by_trunk.items()])
        for source, trunks in targets.items():
            if len(trunks) > 1:
                raise BiletoError('Too many merge targets: {}'.format(trunks))
        return cls.all_mps

    @staticmethod
    def stringify(merges_by_source):
        """Reverse prepare_all() back to strings."""
        urls = ''
        for source_name, merges in sorted(merges_by_source.items()):
            urls += N.join(merge.web_link for merge in merges) + NN
        return urls.strip()

    def __init__(self, *args):
        """Get just the merges relevant to this source package."""
        super().__init__(*args)
        self.merges = self.all_mps[self.name]
        self.target = self.merges[0].target_branch.display_name

    @property
    @lru_cache()
    def lp_branch(self):
        """Identify where on launchpad we store our branch."""
        return 'lp:~/{}/{}-{}-{}-{}'.format(
            self.merges[0].target_branch.project.name,
            self.name,  # Source package name
            self.distro,
            self.series.name,
            self.ppa.name)

    def create_version(self):
        """Identify the highest version and then increment from there."""
        versions = []
        for branch in (self.target, self.lp_branch, 'WORKDIR'):
            with suppress(VCSError):
                result = yield from VCS.source_version(branch, **self.env)
                versions.append(result)
        with suppress(AttributeError):
            versions.append(self.get_source(self.ppa).source_package_version)
        new_version = V(V.max(*versions)).create_new(self.series.version)
        log.info('Using version %s for %s.', new_version, self.name)
        self.env.update(VERSION=new_version)

    @coroutine
    def collect(self):
        """Branch trunk and merge in all merges."""
        yield from VCS.branch(self.target, **self.env)
        changelog = defaultdict(list)
        for merge in self.merges:
            message = (merge.commit_message or N).splitlines()[0]
            bugs = format_bugs(merge.source_branch.linked_bugs)
            authors = yield from VCS.merge(
                merge.source_branch.web_link,
                MESSAGE=message,
                BUGS=bugs,
                **self.env)
            if message:
                changelog[authors.strip(COMMASPACE)].append(message + bugs)
        yield from self.create_version()
        for authors, messages in sorted(changelog.items()):
            names, emails = parse_authors(authors)
            for message in messages:
                yield from VCS.append_changelog(
                    self.target,
                    DEBFULLNAME=names or environ['BLAME_FULLNAME'],
                    DEBEMAIL=emails or environ['BLAME_EMAIL'],
                    MESSAGE=message,
                    **self.env)
        yield from VCS.tag_release(self.target, **self.env)

    @coroutine
    def expose(self):
        """Push our locally built branch to launchpad."""
        yield from VCS.expose(self.lp_branch, **self.env)
