"""Bileto Package Base Class

Defines steps common to all types of packages that we handle.
"""

from os import environ
from os.path import join
from contextlib import suppress
from asyncio import coroutine, get_event_loop

from bileto.settings import Config

from bileto.worker.log import log
from bileto.worker.vcs import VCS


class Package:
    """Represent the steps necessary to build & publish a package."""
    instances = {}

    def __init__(self, name, distro, series, dest, ppa):
        self.distro = distro
        self.series = series
        self.name = name
        self.dest = dest
        self.ppa = ppa
        self.env = dict(
            SOURCE=name,
            SERIES=series.name,
        )
        self.instances[join(series.name, name)] = self

    @coroutine
    def validate(self):
        """NOP; Validation handled globally."""

    @staticmethod
    def post_validate(ticket):
        """Report ticket summary."""
        ticket.validate()
        log.info(ticket.summarize())

    @coroutine
    def build(self):
        """Build source package."""
        yield from VCS.build(**self.env)

    @coroutine
    def upload(self):
        """Upload source package to PPA."""
        yield from VCS.upload(
            DISTRO=self.distro,
            PPA_NAME=self.ppa.name,
            PPA_TEAM=Config.ppa_team,
            **self.env
        )

    @coroutine
    def expose(self):
        """NOP; Subclasses may optionally implement this."""

    def get_source(self, archive):
        """Get the lplib object representing this source."""
        with suppress(IndexError):
            return archive.getPublishedSources(
                distro_series=self.series,
                source_name=self.name,
                order_by_date=True,
                exact_match=True,
            )[0]

    def find_dsc_url(self, source):
        """Identify the URL to a DSC file for this package in an archive."""
        if source:
            for url in source.sourceFileUrls():
                if url.endswith('.dsc'):
                    return url
        return ''

    @coroutine
    def diff(self):
        """Compare this package to the version in the destination archive."""
        our_source = self.get_source(self.ppa)
        our_dsc = self.find_dsc_url(our_source)
        yield from VCS.diff_prefetch(DSC_URL=our_dsc, **self.env)
        dest_source = self.get_source(self.dest)
        self.env.update(
            COMPONENT=dest_source.component_name if dest_source else 'NEW',
        )
        dest_dsc = self.find_dsc_url(dest_source or self.get_source(
            self.dest.distribution.main_archive))
        yield from VCS.diff_dest(DSC_URL=dest_dsc, **self.env)

    @staticmethod
    def post_diff(ticket):
        """Upload all diffs to swift after they are all generated."""
        environ.update(
            OS_REGION_NAME=Config.os_region_name,
            OS_TENANT_NAME=Config.os_tenant_name,
            OS_AUTH_URL=Config.os_auth_url,
            OS_USERNAME=Config.os_username,
            OS_PASSWORD=Config.os_password,
        )
        stdout = get_event_loop().run_until_complete(VCS.swift_upload(
            # Don't put creds here as they'll leak into debug logs.
            REQUESTID=str(ticket.request_id),
        ))
        log.info('Uploaded diffs:\n%s', stdout)
        ticket.artifacts = stdout
        ticket.commit()

    @staticmethod
    def post_abandon(ticket):
        """Deactivate ticket."""
        ticket.set_status('Abandoned')
