"""Bileto Version Control System

This is just a trivial pass-through that shells out to vcs.sh in order to
perform various vcs-related actions. All VCS knowledge should ideally be
encapsulated in vcs.sh in order to easily add support for new VCSs later.
"""

from os import environ
from os.path import join
from asyncio.subprocess import PIPE
from asyncio import create_subprocess_exec, gather, get_event_loop

from bileto.worker.log import log
from bileto.settings import Config
from bileto.models import BiletoError


def parallelize(tasks):
    """Do each task in parallel with asyncio."""
    get_event_loop().run_until_complete(gather(*tasks))


class VCSError(BiletoError):
    """Error raised when VCS operation fails."""


class VCS:
    """Handle shelling out to subprocesses in parallel using coroutines."""
    script = join(Config.ROOT, 'scripts', 'vcs.sh')

    def __getattr__(self, attr):
        """Proxy method calls down to vcs.sh."""
        def passthrough(*args, **kwargs):
            """Pass args down to vcs.sh."""
            kwargs.update(ACTION=attr)
            kwargs.setdefault('SERIES', '')
            kwargs.setdefault('SOURCE', '')
            env = dict(environ)
            env.update(**kwargs)
            proc = yield from create_subprocess_exec(
                self.script, *args, stdout=PIPE, stderr=PIPE, env=env)
            log.debug('Forked: %s %s %s', kwargs, self.script, args)
            stdout, stderr = yield from proc.communicate()
            stderr = stderr.decode('utf-8', errors='replace').strip()
            stdout = stdout.decode('utf-8', errors='replace').strip()
            if stderr:
                log.info('\n{}\n'.format(stderr))
            if proc.returncode != 0:
                raise VCSError(stdout)
            return stdout
        return passthrough

VCS = VCS()  # Single global instance
