#!/usr/bin/env python3

"""Britney runner.

This script iterates over all appropriate bileto requests, downloads archive
indexes, and runs britney with those indexes as inputs.
"""

import sys

from glob import glob
from json import dumps
from time import strftime
from pprint import pformat
from datetime import datetime
from contextlib import suppress
from subprocess import Popen, PIPE
from fcntl import LOCK_EX, LOCK_NB, flock

from os import unlink, symlink
from os.path import abspath, dirname, exists

from requests import get, post

HOME = dirname(abspath(__file__))
ROOT = dirname(HOME)
sys.path.append(ROOT)

# This import doesn't work without hacking sys.path.
# pylint: disable=wrong-import-position
from bileto.settings import Config  # noqa


PUBLIC = '/static/britney/'
OUTPUT = '/tmp/britney_output/'

LOG_NAME = OUTPUT + strftime('log_%Y%m%d_%H%M%S.txt')

API_PREFIX = 'http://127.0.0.1:8080/v1/tickets?'
READ_API = API_PREFIX + 'lander_signoff=Approve&siloname=land&status=Success'
WRITE_API = API_PREFIX + 'token='
FETCH = HOME + '/fetch-indexes'
EXCUSES = OUTPUT + 'ticket-{}/*/excuses.html'
YAML = OUTPUT + 'ticket-{}/*/excuses.yaml'
SLASH = '/'
PLUS = '+'
SPACE = ' '
EMPTY = ''
BNEWLINE = b'\n'
NEWLINE = '\n'
REQS = {}


# TODO: Consider replacing this with a standard locking module
class Lock:
    """An exclusive file lock."""

    def __init__(self, lockfile):
        self.lockfile = lockfile
        self.lock_fd = None

    def lock(self):
        """Acquire the lock."""
        if self.lock_fd is None:
            self.lock_fd = open(self.lockfile, 'w')
        flock(self.lock_fd, LOCK_EX | LOCK_NB)

    def unlock(self):
        """Release the lock."""
        if self.lock_fd is not None:
            self.lock_fd.close()
            self.lock_fd = None


def cleanup(lock):
    """Actions that must be run no matter how badly everything exploded."""
    # update the last-run.txt symlink
    if exists(LOG_NAME):
        last_run_link = OUTPUT + 'last-run.txt'
        with suppress(FileNotFoundError):
            unlink(last_run_link)
        symlink(LOG_NAME, last_run_link)
    lock.unlock()


def write(data):
    """Write unbuffered stdout."""
    sys.stdout.buffer.write(data.encode('utf-8'))


def get_ppa(siloname):
    """Extract just the ppa name from the siloname."""
    return siloname[siloname.index(SLASH) + 1:]


def fetch_indexes(*args):
    """Invoke the index-fetcher script with the given args."""
    return Popen((FETCH,) + args, stdout=PIPE, stderr=PIPE)


def get_requests():
    """Return list of all active requests from Bileto."""
    resp = get(READ_API).json()
    tickets = resp.get('requests', [])
    # Skip pre-approved tickets to save time.
    return [
        t for t in tickets
        if 'Approved' not in (t['britney_signoff'], t['qa_signoff'])
    ]


def read_yaml_status(request_id):
    """Distill britney status into single word."""
    raw = ''
    for filename in glob(YAML.format(request_id)):
        with open(filename, 'r', encoding='utf-8') as data:
            raw += data.read()
    if 'is-candidate: false' in raw:
        if 'REGRESSION' in raw:
            return 'Failed'
        if 'RUNNING\n' in raw:
            return 'Running'
        return 'Failed'
    if 'is-candidate: true' in raw:
        return 'Approved'
    if raw:
        return 'N/A'
    return ''


def update_request(req):
    """Send excuses links to bileto."""
    request_id = req['request_id']
    files = sorted(glob(EXCUSES.format(request_id)))
    if files:
        new_req = dict(
            author='britney-bot',
            request_id=request_id,
            britney_signoff=read_yaml_status(request_id),
            autopkgtest=NEWLINE.join(
                name.replace(OUTPUT, PUBLIC) for name in files))
        write(pformat(new_req))
        resp = post(
            WRITE_API + Config.api_token, dumps(new_req, sort_keys=True),
            headers={'content-type': 'application/json'})
        write('\n{} Updating request {}: {} {}\n\n'.format(
            datetime.utcnow(), request_id, resp.status_code, resp.reason))


def build_siloname_dict(reqs):
    """Build a dict mapping series to silonames."""
    runs = {}
    all_series = set()
    for req in reqs:
        landing = get_ppa(req['siloname'])
        REQS[landing] = req
        serieses = req['series'].split(PLUS)
        all_series.update(serieses)
        runs[landing] = sorted(serieses)
    return all_series, runs


def parallelize_fetching(all_series, runs):
    """Invoke fetchers in parallel but serialize output on stdout."""
    procs = []
    # Parallelize most of the fetching...
    for series in sorted(all_series):
        procs.append(fetch_indexes(series))
    # ... serialize the output ...
    for proc in procs:
        print_proc(proc)
    # ... and serialize the britney invocations
    for siloname, serieses in sorted(runs.items()):
        ticket = REQS[siloname]
        requestid = str(ticket['request_id'])
        for series in serieses:
            print_proc(fetch_indexes(series, siloname, requestid))
        update_request(ticket)


def print_proc(proc):
    """Print the output from a subprocess."""
    sys.stdout.buffer.write(BNEWLINE.join(
        (b'', SPACE.join(proc.args).encode('utf-8')) + proc.communicate()))


def main():
    """Handle errors."""
    start = datetime.utcnow()
    lock = Lock(OUTPUT + 'lock')
    try:
        lock.lock()
        sys.stderr = sys.stdout = open(LOG_NAME, 'x')
        reqs = get_requests()
        parallelize_fetching(*build_siloname_dict(reqs))
    except BlockingIOError:
        return 0
    except (IOError, ConnectionRefusedError):
        write('Failed to connect to Bileto. Is it running?')
    else:
        end = datetime.utcnow()
        write('\n{} Completed in {}.'.format(end, end - start))
    finally:
        cleanup(lock)
    return 0


if __name__ == '__main__':                          # pragma: no cover
    sys.exit(main())
