#!/bin/sh
# The goal of this script is to define a subset of VCS commands needed by
# Bileto, in a modular way, such that it is then easy to add git support
# in the near future. For now, think of this as an abstraction layer on top
# of bzr so that the python code in Bileto doesn't need to have any bzr
# knowledge at all.

# Note: All informational logging should be routed to STDERR. STDOUT becomes
# the return value of this script in the calling python code, and must be kept
# clean for parsing. Unless the exit code is nonzero, then STDOUT should be
# a concise error message. See vcs.py for how STDOUT/STDERR are treated.

# Fuzz startup time a bit since this gets forked in parallel a lot.
sleep "$(shuf --input-range 0-10 --head-count 1)"

set -eu

TWENTY_MINUTES="1200"
CACHE_DIR="/tmp/bileto/vcs-cache"
WORKSPACE="$(pwd)"
WORKPARENT="$WORKSPACE/$SERIES/$SOURCE"
WORKDIR="$WORKPARENT/$SOURCE"
[ -d "$WORKDIR" ] && cd "$WORKDIR" || true
IFS='
'


if [ $# -gt 0 ]; then
    BRANCH="$1"
    CACHE_BASE="$(echo "$BRANCH" | perl -lpe 's/([^-_a-z0-9:~])/sprintf("%%%02X", ord($1))/gie')"
    CACHE="$CACHE_DIR/$CACHE_BASE"
    CACHE_SOURCE="$WORKPARENT/$CACHE_BASE"
fi


mkdir --parents "$CACHE_DIR" "$WORKPARENT" 1>&2


#####################
# Utility Functions #
#####################

die() {
    echo "$SERIES/$SOURCE:" "$@" 2>&1
    exit 1
}

die_gently() {
    echo "$SERIES/$SOURCE:" "$@" 1>&2
    exit 0
}

loudly() {
    # Echo a command before running it, and pipe everything to stderr
    echo + "$@" 1>&2
    "$@" 1>&2
}

stdout() {
    # Echo a command before running it, but keep it on stdout for processing
    echo + "$@" 1>&2
    "$@" 2>&1
}

parsechangelog() {
    (cd "$1" && dpkg-parsechangelog --show-field "$2"     \
             || die "Failed to parse changelog in $1")
}

force_changelog_version() {
    loudly sed --in-place --expression "1s/.*/$SOURCE ($VERSION) $SERIES; urgency=medium/g" debian/changelog || die "Failed to set changelog version to $VERSION."
}

gpg_key() {
    # Sort gpg secret keys by creation date and pick the newest one.
    gpg --list-secret-keys --with-colons | awk -F: '/^sec/ {print $6 "\t" $5}' | sort | tail -1 | cut --field 2
}

download_from_dsc() {
    loudly dget --quiet --allow-unauthenticated --download-only "$1" || die "Failed to download DSC file $1"
    DSC_FILE="$(basename "$1")"
    loudly gpg --keyserver "keyserver.ubuntu.com" --keyserver-options "auto-key-retrieve" --verify "$DSC_FILE" || die "Failed to verify DSC file $DSC_FILE"
}

debdiff_wrapper() {
    # Disable gpgv because we already verified the DSCs manually.
    ln --symbolic --force /bin/true $HOME/gpgv
    debdiff_stderr="$(mktemp debdiff-stderr-XXX)"
    (export PATH=$HOME:$PATH && set -x && debdiff --diffstat "$1" "$2" 1>"$3" 2>"$debdiff_stderr") || true
    cat "$debdiff_stderr" 1>&2
    if egrep --ignore-case --quiet 'error|fail' "$debdiff_stderr"; then
        die "debdiff failed: see log for details."
    fi
}

die_if_timed_out() {
    # shellcheck disable=SC2046
    if [ $(ps --pid "$$" --format "etimes=") -gt "$TWENTY_MINUTES" ]; then
        die "Timed out."
    fi
}

get_from_primary() {
    printf "Waiting for %s to finish " "$1" 1>&2
    until [ -f "$1/../done" ]; do
        die_if_timed_out
        printf "." 1>&2
        sleep 10
    done
    printf " done!\n" 1>&2
    loudly cp --archive "$1" "$WORKDIR" || die "Failed to copy source tree from primary."
    cd "$WORKDIR"
}

run_hook() {
    scripts="$(dirname "$(readlink --canonicalize "$0")")"
    export HOOK="$1"
    chmod 711 "$WORKSPACE"  # Needed so unity8 hook script can cd into ./po/
    (cd "$WORKDIR" && sudo --preserve-env /bin/sh $scripts/run_hook.sh || die "Failed to run hook $HOOK.")
}


####################
# Generic Commands #
####################

do_append_changelog() {
    loudly dch --multimaint-merge --release-heuristic changelog --newversion "$VERSION" "$MESSAGE" || die "Failed to add changelog message."
}

do_get_secondary() {
    get_from_primary "$WORKSPACE/$PRIMARY/$SOURCE/$SOURCE"
    orig_version="$(parsechangelog "$WORKDIR" Version)"
    VERSION="$(echo "$orig_version" | sed "s/\+\([[:digit:]]\{2\}\.\)\{2\}\([[:digit:]]\{8\}\)/+$SERIES_VERSION.\2/")"
    force_changelog_version
    run_hook pre_release_hook
    loudly touch "$WORKPARENT/done"
}

do_get_nongles() {
    nongles="$(echo $SOURCE | sed -n 's/-gles$//p')"
    get_from_primary "$WORKSPACE/$SERIES/$nongles/$nongles"
    run_hook convert_to_gles
    new_name="$(parsechangelog "$WORKDIR" Source)"
    [ "$new_name" = "$SOURCE" ] || die "Found $new_name in debian/changelog but expected $SOURCE"
    loudly touch "$WORKPARENT/done"
}

do_build() {
    upstream="$(dpkg-parsechangelog --show-field Version | sed -e 's/^[^:]\+://' -e 's/-[^-]\+$//')"
    orig_tar="../${SOURCE}_${upstream}.orig.tar.gz"
    loudly tar --create --auto-compress --file "$orig_tar" \
        --exclude='.bzr*' --exclude='.git*' --exclude="debian" . || die "Failed to create orig.tar."

    # -sa: Force inclusion of original source.
    # -d:  Do not check build dependencies and conflicts.
    # -S:  Only build source package, no binaries.
    # -i:  Ignore bzr/git files from generated diff (can only be specified once).
    # -I:  Ignore VCS files from generated tarball, plus .bzr-builddeb which isn't included in default -I
    # -nc: Do not run `./debian/rules clean`
    # -k:  Sign .changes with this key.
    loudly dpkg-buildpackage -sa -d -S -i'(^|/)(\.bzr|\.git)' -I -I'.bzr*' -nc -k"$(gpg_key)" --force-sign || die "Failed to build source package."
}

do_upload() {
    loudly dput ppa:$PPA_TEAM/$DISTRO/$PPA_NAME $WORKPARENT/${SOURCE}_*.changes || die "Failed to upload package."
}

do_diff_prefetch() {
    cd "$WORKPARENT"
    our_dsc="$(echo ${SOURCE}_*.dsc)"
    if [ ! -f "$our_dsc" ]; then
        [ -z "$DSC_URL" ] || download_from_dsc "$DSC_URL"
    fi
}

_build_packaging_diff() {
    printf " * This diff only shows the debian/ directory! It is not a full content diff!\n\n"
    [ -n "$DSC_URL" ] || printf " * Source package must be preNEWed by an archive admin because it has never been released in Ubuntu before.\n\n"
    [ -z "$COMPONENT" ] || printf " * Source package is a %s component.\n\n" "$COMPONENT"
    packages="$(sed -n 's/^[+-]Package: //p' <"$2" | sort | uniq --unique)"
    [ -z "$packages" ] || printf " * Please consult an archive admin about adding or removing these packages: \n%s\n\n" "$packages"
    stdout filterdiff --clean --include '*debian/*' <"$1" || die "Failed to build packaging diff."
}

do_diff_dest() {
    cd "$WORKPARENT"
    our_dsc="$(echo ${SOURCE}_*.dsc)"
    [ -f "$our_dsc" ] || die_gently "Diff failed: No DSC found, was it ever built?"
    content="${SERIES}_${SOURCE}_content.diff"
    packaging="${SERIES}_${SOURCE}_packaging_changes.diff"
    raw="$packaging.raw"
    touch "$content"
    if [ -n "$DSC_URL" ]; then
        download_from_dsc "$DSC_URL"
        dest_dsc="$(basename "$DSC_URL")"
        debdiff_wrapper "$dest_dsc" "$our_dsc" "$content"
        stdout filterdiff --clean --include '*debian/*' --exclude '*changelog' <"$content" >"$raw" || die "Failed to discover packaging changes."
    else
        echo '---NEW SOURCE PACKAGE---' >"$raw"
    fi
    if [ -s "$raw" ]; then
        _build_packaging_diff "$content" "$raw" >"$packaging"
    fi
}

do_swift_upload() {
    cd "$WORKSPACE"
    retry="Please try regenerating diffs."
    root="https://objectstorage.prodstack4-5.canonical.com/v1"
    auth="$(swift stat -v | sed -n 's/^\s\+Account:\s\+//p')"
    timestamp="$(date --utc '+%Y-%m-%d_%H:%M:%S')"
    container="bileto-$REQUESTID"
    [ -n "$auth" ] || die "Failed to determine swift account. $retry"

    diffs=""
    for filename in */*/*.diff; do
        base="$(basename "$filename")"
        confirmed="$(swift upload "$container" "$filename" --object-name="$timestamp/$base")" || die "Failed to upload diffs. $retry"
        size="$(wc --lines $filename | awk '{ print $1 }')"
        diffs="$diffs\n$root/$auth/$container/$confirmed ($size lines)"
    done

    swift post "$container" --read-acl '.r:*,.rlistings' || die "Failed to publish diffs publicly. $retry"
    printf "$diffs"
}


################
# BZR Commands #
################

bzr_do_fill_local_cache() {
    [ -d "$CACHE" ] || loudly bzr branch "$BRANCH" "$CACHE" || die "Failed to branch $BRANCH"
    loudly bzr pull --directory "$CACHE" --overwrite || die "Failed to update $BRANCH"
    touch "$CACHE"
}

bzr_do_source_name() {
    bzr_do_fill_local_cache
    parsechangelog "$CACHE" Source
}

bzr_do_source_version() {
    [ -d "$CACHE" ] && where="$CACHE" || where="$WORKDIR"
    parsechangelog "$where" Version
}

bzr_do_branch() {
    loudly bzr init-repo "$WORKPARENT" || die "Failed to initialize bzr repo."
    loudly bzr branch "$CACHE" "$WORKDIR" || die "Failed to branch trunk."
}

bzr_do_merge() {
    authors=""
    output=""
    loudly bzr branch "$BRANCH" "$CACHE_SOURCE" || die "Failed to branch $BRANCH"
    missing="$CACHE_SOURCE.missing"
    stdout bzr missing --theirs "$CACHE_SOURCE" >"$missing"
    for author in $(sed -n 's/^committer: //p' <"$missing" | sort | uniq); do
        output="$author, $output"
        authors="$authors--author=$(echo $author | sed 's/ /\ /g') "
    done
    loudly bzr merge "$CACHE_SOURCE" || die "Failed to merge $BRANCH"
    if [ -n "$MESSAGE" ]; then
        loudly bzr commit --unchanged $authors --message "$MESSAGE$BUGS" || die "Failed to commit $BRANCH"
    else
        loudly debcommit || die "Failed to debcommit $BRANCH. You must supply either an MP commit message or a custom debian/changelog entry."
    fi
    echo "$output"
    rm --recursive --force "$CACHE_SOURCE"
}

bzr_do_tag_release() {
    find debian/ -type f -print0 | xargs -0 perl -p -i -e "s/0replaceme/$VERSION/g"
    # Call dch to correctly set the blame / timestamp at the end of the changelog entry
    loudly dch --release --force-distribution --distribution "$SERIES" "" || die "Failed to mark debian/changelog for release."
    # The version won't have been set correctly if the only MP used debcommit though, so fix that
    force_changelog_version
    run_hook pre_release_hook
    loudly bzr commit --unchanged --message "Releasing $VERSION" || die "Failed to commit release."
    loudly bzr tag --force "$VERSION" || die "Failed to tag release."
    loudly touch "$WORKPARENT/done"
}

bzr_do_expose() {
    loudly bzr push "$BRANCH" --overwrite || die "Failed to push source tree to launchpad."
}


################
# GIT Commands #
################

git_do_source_name() {
    die "Not yet implemented"
    parsechangelog "$CACHE" Source
}

git_do_source_version() {
    parsechangelog "$CACHE" Version
}


ACTION="do_$ACTION"
VCS="bzr"  # TODO: Implement git & detect when git is in use.
VCS_ACTION="${VCS}_$ACTION"

# shellcheck disable=SC2039
type "$VCS_ACTION" 2>/dev/null 1>&2 && $VCS_ACTION "$@" || $ACTION "$@"
