"""Britney tests."""

from json import loads
from contextlib import ExitStack
from collections import defaultdict
from multiprocessing import Process, Queue
from os.path import dirname, islink, join, realpath
from tempfile import NamedTemporaryFile, TemporaryDirectory
from unittest.mock import MagicMock, call, patch
from unittest import TestCase

from britney import iterate

import tests
DATA_DIR = join(dirname(tests.__file__), 'data')
COLON = ':'

# Read the tickets.json sample data file for use in mocking requests.get()
with open(join(DATA_DIR, 'tickets.json'), 'r', encoding='utf-8') as ticketsfp:
    TICKETS_JSON = loads(ticketsfp.read())


# Don't try to cover this function since it only executes in a subprocess.
def get_lock_and_fail(lockfile, queue):                     # pragma: no cover
    """Fail to get the lock in a subprocess."""
    try:
        iterate.Lock(lockfile).lock()
    except BlockingIOError:
        queue.put(True)
    else:
        queue.put(False)


class TestBritneyIterate(TestCase):
    """Test the britney/iterate.py script."""

    maxDiff = None

    def test_lock(self):
        """Test that the lock is exclusive."""
        with NamedTemporaryFile() as lockfile:
            # Acquire the lock in the foreground process
            lock = iterate.Lock(lockfile.name)
            lock.lock()
            self.addCleanup(lock.unlock)
            # Run a subprocess that tries to acquire the same lock.  This
            # should block until the foreground gives up the lock.
            q = Queue()
            p = Process(target=get_lock_and_fail, args=(lockfile.name, q))
            self.addCleanup(p.join)
            p.start()
            # The subprocess should have gotten a BlockingIOError.
            self.assertTrue(q.get())

    def test_cleanup(self):
        """Cleanup actions must update the last-run symlink."""
        with ExitStack() as resources:
            log_file = resources.enter_context(NamedTemporaryFile())
            output_file = resources.enter_context(NamedTemporaryFile())
            resources.enter_context(
                patch.object(iterate, 'LOG_NAME', log_file.name))
            resources.enter_context(
                patch.object(iterate, 'OUTPUT', output_file.name))
            lock = MagicMock()
            iterate.cleanup(lock)
            self.assertTrue(lock.unlock.called)
            link_file = output_file.name + 'last-run.txt'
            self.assertTrue(islink(link_file))
            self.assertEqual(realpath(link_file), log_file.name)

    @staticmethod
    def _requests():
        with patch.object(iterate, 'get') as mock:
            mock.return_value.json.return_value = TICKETS_JSON
            return iterate.get_requests()

    def test_request(self):
        """Return the list of request."""
        requests = self._requests()
        self.assertIsInstance(requests, list)
        self.assertEqual(len(requests), 2)
        self.assertEqual([req['creator'] for req in requests],
                         ['renatofilho', 'cjwatson'])

    def test_get_ppa(self):
        """Dig the silo name out of the requests."""
        requests = self._requests()
        self.assertEqual(
            [iterate.get_ppa(req['siloname']) for req in requests],
            ['landing-029', 'landing-046'])

    def test_build_silo_name(self):
        """Test the series -> silonames dictionary."""
        requests = self._requests()
        self.assertSequenceEqual(
            iterate.build_siloname_dict(requests), ({
                'xenial', 'vivid',
            }, {
                'landing-029': ['vivid', 'xenial'],
                'landing-046': ['vivid', 'xenial'],
            }))

    def test_read_yaml_status_regression(self):
        "Regression status."""
        YAML = join(DATA_DIR, 'regression', 'britney', '*/{}/excuses.yaml')
        with patch.object(iterate, 'YAML', YAML):
            status = iterate.read_yaml_status('landing-041')
        self.assertEqual(status, 'Failed')

    def test_read_yaml_status_running(self):
        """Running status."""
        YAML = join(DATA_DIR, 'running', 'britney', '*/{}/excuses.yaml')
        with patch.object(iterate, 'YAML', YAML):
            status = iterate.read_yaml_status('landing-041')
        self.assertEqual(status, 'Running')

    def test_read_yaml_status_not_considered(self):
        """Not considered status."""
        YAML = join(DATA_DIR, 'not-considered', 'britney', '*/{}/excuses.yaml')
        with patch.object(iterate, 'YAML', YAML):
            status = iterate.read_yaml_status('landing-041')
        self.assertEqual(status, 'Failed')

    def test_read_yaml_status_regress_override(self):
        """Autopkgtest regressions, but overridden."""
        YAML = join(DATA_DIR, 'overrides', 'britney', '*/{}/excuses.yaml')
        with patch.object(iterate, 'YAML', YAML):
            status = iterate.read_yaml_status('landing-023')
        self.assertEqual(status, 'Approved')

    def test_read_yaml_status_valid_candidate(self):
        """Valid candidate status."""
        YAML = join(DATA_DIR, 'valid-candidate', 'britney',
                    '*/{}/excuses.yaml')
        with patch.object(iterate, 'YAML', YAML):
            status = iterate.read_yaml_status('landing-041')
        self.assertEqual(status, 'Approved')

    def test_read_yaml_status_na(self):
        """Some indeterminate status."""
        YAML = join(DATA_DIR, 'na', 'britney', '*/{}/excuses.yaml')
        with patch.object(iterate, 'YAML', YAML):
            status = iterate.read_yaml_status('landing-041')
        self.assertEqual(status, 'N/A')

    def test_read_yaml_status_none_of_the_above(self):
        """No status."""
        YAML = join(DATA_DIR, 'nota', 'britney', '*/{}/excuses.yaml')
        with patch.object(iterate, 'YAML', YAML):
            status = iterate.read_yaml_status('landing-041')
        self.assertEqual(status, '')

    def test_parallelize_fetching(self):
        """Fetch, but don't actually shell out to fetch-indexes."""
        mocks = defaultdict(MagicMock)
        with ExitStack() as resources:
            resources.enter_context(patch.object(iterate, 'Popen'))
            iterate.Popen.side_effect = lambda args, **ignore: (
                setattr(mocks[args], 'args', ('fetch-indexes',) + args[1:]),
                setattr(mocks[args].communicate, 'return_value', (b'OUTPUT',)),
                mocks[args],
            )[-1]
            # Capture stdout to evaluate it later.  We do it this way to be
            # compatible with the API expected by print_proc().
            log_file = resources.enter_context(NamedTemporaryFile())
            stdout = MagicMock()
            stdout.buffer = log_file
            resources.enter_context(patch('sys.stdout', stdout))
            resources.enter_context(patch.object(iterate, 'update_request'))
            iterate.REQS['landing-001'] = MagicMock()
            iterate.REQS['landing-000'] = MagicMock()
            iterate.REQS['landing-000'].__getitem__.return_value = '9000'
            iterate.REQS['landing-001'].__getitem__.return_value = '5000'
            iterate.parallelize_fetching({'xenial', 'vivid'}, {
                'landing-000': ['xenial'],
                'landing-001': ['vivid'],
            })
            self.assertSequenceEqual(iterate.update_request.mock_calls, [
                call(iterate.REQS['landing-000']),
                call(iterate.REQS['landing-001']),
            ])
            # Closing the temporary file will delete it.
            log_file.seek(0)
            output = log_file.read().decode('utf-8')
            self.assertMultiLineEqual(output, """
fetch-indexes vivid
OUTPUT
fetch-indexes xenial
OUTPUT
fetch-indexes xenial landing-000 9000
OUTPUT
fetch-indexes vivid landing-001 5000
OUTPUT""")

    def test_main(self):
        """Now for the big-ol' test of main."""
        with ExitStack() as resources:
            # Mock out a bunch of stuff already tested.
            resources.enter_context(
                patch.object(iterate, 'Config', api_token='token'))
            resources.enter_context(patch.object(iterate, 'Lock'))
            # In this case, because main() opens the LOG_NAME exclusively, the
            # file cannot yet exist.  Name a dummy file in a new temporary
            # directory.
            tempdir = resources.enter_context(TemporaryDirectory())
            resources.enter_context(
                patch.object(iterate, 'LOG_NAME', join(tempdir, 'main.log')))
            output_file = resources.enter_context(NamedTemporaryFile())
            resources.enter_context(
                patch.object(iterate, 'OUTPUT', output_file.name))
            mock = resources.enter_context(patch.object(iterate, 'get'))
            mock().json.return_value = TICKETS_JSON
            resources.enter_context(
                patch.object(iterate, 'parallelize_fetching'))
            resources.enter_context(
                patch.object(iterate, 'EXCUSES',
                             join(DATA_DIR, 'excuses', '*/{}/excuses.html')))
            code = iterate.main()
            self.assertEqual(code, 0)

    def test_main_blocking_io_error(self):
        """Cover the BlockIOError stanza."""
        with ExitStack() as resources:
            # Mock out a bunch of stuff already tested.
            output_file = resources.enter_context(NamedTemporaryFile())
            resources.enter_context(
                patch.object(iterate, 'OUTPUT', output_file.name))
            resources.enter_context(
                patch.object(iterate, 'Config', api_token='token'))
            # E.g. the lock can't be acquired.
            resources.enter_context(patch.object(iterate.Lock, 'lock',
                                                 side_effect=BlockingIOError))
            self.assertEqual(iterate.main(), 0)

    def test_main_io_error(self):
        """Cover the IOError/ConnectionRefusedError stanza."""
        with ExitStack() as resources:
            # Mock out a bunch of stuff already tested.
            output_file = resources.enter_context(NamedTemporaryFile())
            resources.enter_context(
                patch.object(iterate, 'OUTPUT', output_file.name))
            resources.enter_context(
                patch.object(iterate, 'Config', api_token='token'))
            resources.enter_context(patch.object(iterate, 'Lock'))
            # In this case, because main() opens the LOG_NAME exclusively, the
            # file cannot yet exist.  Name a dummy file in a new temporary
            # directory.
            tempdir = resources.enter_context(TemporaryDirectory())
            resources.enter_context(
                patch.object(iterate, 'LOG_NAME', join(tempdir, 'main.log')))
            # E.g. get_requests() fails.
            resources.enter_context(patch.object(iterate, 'get_requests',
                                                 side_effect=IOError))
            self.assertEqual(iterate.main(), 0)

    @patch.object(iterate, 'post')
    @patch.object(iterate, 'glob')
    @patch.object(iterate, 'Config')
    @patch.object(iterate, 'read_yaml_status')
    @patch.object(iterate, 'write', MagicMock())
    def test_update_request(self, yaml_mock, conf_mock, glob_mock, post_mock):
        """Send the correct information to bileto."""
        conf_mock.api_token = 'joken'
        yaml_mock.return_value = 'Running'
        glob_mock.return_value = ['excuses.html']
        req = dict(request_id=500, siloname='ubuntu/landing-005')
        self.assertIsNone(iterate.update_request(req))
        post_mock.assert_called_once_with(
            'http://127.0.0.1:8080/v1/tickets?token=joken',
            '{"author": "britney-bot", "autopkgtest": "excuses.html", '
            '"britney_signoff": "Running", "request_id": 500}',
            headers={'content-type': 'application/json'})
