"""Bileto Login Tests

Test all things related to OpenID logins.
"""

from mock import patch

from tests.tests import BiletoTestCase

from bileto.login import identify

SESSION = {}


class LoginTestCase(BiletoTestCase):
    """Test OpenID Logins"""

    def test_login(self):
        """Ensure correct redirect when initiating login."""
        ret = self.app.post('/login', data=dict(
            openid='test',
            next='/',
        ), follow_redirects=False)
        self.assertIn(b'https://login.ubuntu.com/+openid', ret.data)
        self.assertEqual(ret.status_code, 302)

    def test_login_get(self):
        """Ensure login endpoint accepts GET requests as per SSO spec."""
        ret = self.app.get('/login', follow_redirects=False)
        self.assertIn(b'<a href="/">/</a>.', ret.data)
        self.assertEqual(ret.status_code, 302)

    def test_logged_already(self):
        """Ensure correct redirect when already logged in."""
        with self.app.session_transaction() as sesh:
            sesh['nickname'] = 'bileto'
        ret = self.app.get('/login', follow_redirects=False)
        self.assertIn(b'>https://localhost/</a>', ret.data)
        self.assertEqual(ret.status_code, 302)

    @patch('bileto.login.lp')
    @patch('bileto.login.oid')
    @patch('bileto.login.session', SESSION)
    def test_identify(self, oid_mock, lp_mock):
        """Ensure OpenID login can be successfully completed."""
        class Resp:
            """Fake OpenID response class."""
            identity_url = 'http://example.com'
            fullname = 'The Dude'
            nickname = 'lebowski'
            email = 'dude@example.com'
        lp_mock.get_irc_nick.return_value = 'irc_nicky'
        oid_mock.get_next_url.return_value = 'https://localhost/'
        ret = identify(Resp)
        self.assertIn(b'>https://localhost/</a>', ret.data)
        for attr in ('identity_url', 'fullname', 'nickname', 'email'):
            self.assertEqual(getattr(Resp, attr), SESSION[attr])
        oid_mock.get_next_url.assert_called_once_with()
        lp_mock.get_irc_nick.assert_called_once_with('lebowski')
        self.assertEqual(ret.status_code, 302)

    def test_logout(self):
        """Ensure logging out correctly clears session."""
        with self.app.session_transaction() as sesh:
            sesh['foo'] = 'bar'
        ret = self.app.get('/logout', follow_redirects=False)
        self.assertIn(b'https://localhost/</a>.', ret.data)
        self.assertEqual(ret.status_code, 302)
        with self.app.session_transaction() as sesh:
            self.assertNotIn('foo', sesh)
