"""Bileto Launchpadlib Tests

Tests for our launchpadlib wrapper & helper methods.
"""

import logging

from unittest.mock import Mock, patch
from os.path import join

from bileto.lplib import lp

from tests.tests import BiletoTestCase


class LaunchpadTestCase(BiletoTestCase):
    """Test lplib."""

    @patch('bileto.lplib.Launchpad')
    def test_new_connection(self, lp_mock):
        """Connect to Launchpad."""
        lp.__class__._real_instance = None
        logging.debug(lp.me)
        logging.debug(lp.people)
        lp_mock.login_with.assert_called_once_with(
            allow_access_levels=['WRITE_PRIVATE'],
            application_name='bileto',
            credentials_file=join(self.tempdir, '.launchpad.credentials'),
            service_root='production',
            version='devel',
        )

    def test_api_root(self):
        """Ensure we can get LP's API root."""
        self.assertEqual(
            lp.load(lp.MP_ROOT + 'hi'), lp._instance.load.return_value)
        lp._instance.load.assert_called_once_with(
            'https://api.launchpad.net/devel/hi')

    @patch('bileto.lplib.Config.ppa_team', 'us')
    def test_ppa_team(self):
        """Return the PPA team object."""
        self.assertEqual(lp.ppa_team, lp.people['us'])

    def test_get_series(self):
        """Ensure that we get the right series from lplib."""
        alpha = Mock()
        alpha.name = 'alpha'
        alpha.active = True
        beta = Mock()
        beta.name = 'beta'
        beta.active = False
        lp.ubuntu.series = [alpha, beta]
        self.assertEqual(lp.active_series(), ['alpha'])

    @patch('bileto.lplib.lp.people', {})
    def test_get_irc_nick(self):
        """Ensure that we get the freenode irc nick from lplib."""
        freenode = Mock()
        freenode.network = 'irc.freenode.net'
        freenode.nickname = 'freenode-nick'
        quakenet = Mock()
        quakenet.network = 'se.quakenet.org'
        quakenet.nickname = 'quakenet-nick'
        lp.people['fred'] = Mock(irc_nicknames=[freenode, quakenet])
        self.assertEqual(lp.get_irc_nick('fred'), 'freenode-nick')

    @patch('bileto.lplib.lp.people', {})
    def test_no_freenode_nick(self):
        """Ensure that we have the lp user when there is no irc nick in lp."""
        quakenet = Mock()
        quakenet.network = 'se.quakenet.org'
        quakenet.nickname = 'quakenet-nick'
        lp.people['fred'] = Mock(irc_nicknames=[quakenet])
        self.assertEqual(lp.get_irc_nick('fred'), 'fred')

    @patch('bileto.lplib.lp.people', {})
    def test_no_irc(self):
        """Ensure that we have the lp user when there is no irc info in lp."""
        lp.people['gary'] = Mock(irc_nicknames=[])
        self.assertEqual(lp.get_irc_nick('gary'), 'gary')

    @patch('bileto.lplib.lp.people', {})
    def test_not_lp_user(self):
        """If the lp user can not be found ensure the username is returned."""
        self.assertEqual(lp.get_irc_nick('bob'), 'bob')

    @patch('bileto.lplib.lp.people', {})
    def test_get_teams(self):
        """Get list of teams for a user."""
        lp.people['jim'] = Mock(memberships_details=[
            Mock(self_link='launchpad/~canonical/+blah'),
            Mock(self_link='launchpad/~ci-train-users/+blah'),
        ])
        self.assertEqual(lp.get_teams('jim'), ['canonical', 'ci-train-users'])
        self.assertEqual(lp.get_teams('bob'), [])
