"""Bileto Models Test

Most of models.py is tested indirectly by test_v1.py. These tests just
cover a few small things missed by that, which should generally only be
bits of code used in the main.py subprocess.
"""

from unittest.mock import patch

from tests.tests import WorkerTestCase

from bileto.models import BiletoError, Request


EMPTY = ''
ENVIRON = dict(
    BLAME='Canada',
    REQUEST_URL_ROOT='http://example.com',
    REQUEST_PATH='/log/50/build/latest',
)


class ModelsTestCase(WorkerTestCase):
    """Test Models Script."""

    @patch('bileto.models.db')
    @patch('bileto.models.Comment')
    @patch('bileto.models.environ', ENVIRON)
    def test_request_set_status(self, comment, db):
        """Set the status easily & correctly."""
        req = Request(request_id=500, siloname='splash/landing-9000')
        req.set_status('hello')
        self.assertEqual(req.status, 'hello')
        comment.assert_called_once_with(
            author='Canada',
            log_url='/log/50/build/latest',
            published_versions='',
            request_id=500,
            siloname='splash/landing-9000',
            text='hello',
        )
        db.session.add.assert_called_once_with(comment.return_value)
        db.session.commit.assert_called_once_with()

    @patch('bileto.models.environ', ENVIRON)
    def test_request_summarize(self):
        """Display a handy summary."""
        req = Request(
            dest=EMPTY,
            landers='jimbo',
            merge_proposals='http://foo/merge',
            request_id=5,
            series='yakkety+xenial+vivid',
            sources='qtmir qtubuntu',
        )
        self.assertEqual(
            req.summarize(),
            'Packages:\n'
            'https://launchpad.net/ubuntu/+source/qtmir\n'
            'https://launchpad.net/ubuntu/+source/qtubuntu\n\n'
            'More info:\n'
            'http://example.com/#/ticket/5\n')
        self.assertEqual(req.merge_proposals_list, ['http://foo/merge'])

    def test_request_validate(self):
        """Prevent invalid requests from being built."""
        req = Request(
            series='xenial+vivid',
            sync_request='anything',
            distribution='wat',
        )
        with self.assertRaisesRegex(BiletoError, 'Syncs are no longer'):
            req.validate()
        req.sync_request = EMPTY
        with self.assertRaisesRegex(BiletoError, 'Need either merges or'):
            req.validate()
        req.sources = 'webbrowser-app'
        with self.assertRaisesRegex(BiletoError, 'not a supported distro'):
            req.validate()
        req.distribution = 'ubuntu'
        with self.assertRaisesRegex(BiletoError, 'migrated to yakkety'):
            req.validate()
        req.series = EMPTY
        req.validate()

    def test_mp_comments(self):
        """Ensure MP list can contain comments, safely ignored."""
        req = Request(merge_proposals="# zing\n#\n# \nhttp://a\n#http://b")
        self.assertEqual(req.merge_proposals_list, ['http://a'])
