"""Bileto Streams Test

Test streaming process output to HTTP clients.
"""

from os import makedirs
from os.path import join

from fcntl import flock, LOCK_EX
from unittest.mock import call, patch

from tests.tests import BiletoTestCase

from bileto.app import db
from bileto.models import Request
from bileto.settings import Config
from bileto import streams

# Unused var to appease pylint
MODULES = [streams]


class StreamsTestCase(BiletoTestCase):
    """Test Streams Interface."""

    def setUp(self):
        """Fake login to easier test authenticated endpoints."""
        super().setUp()
        with self.app.session_transaction() as sesh:
            sesh['teams'] = ['ubuntu-core-dev']
            sesh['nickname'] = 'steve'
            sesh['email'] = 'steve@ste.ve'
            sesh['fullname'] = 'Steven Testson'
        db.session.add(Request(sources='qtmir qtmir-gles'))
        db.session.commit()

    def test_invalid_input(self):
        """Ensure /v1/tickets gives empty result set."""
        ret = self.app.get('/log/five/build/')
        self.assertEqual(ret.status_code, 404)
        ret = self.app.get('/log/1/obliterate/')
        self.assertEqual(ret.status_code, 404)
        ret = self.app.get('/log/my/axe/')
        self.assertEqual(ret.status_code, 404)
        ret = self.app.get('/log/1/diff/')
        self.assertEqual(ret.status_code, 200)
        self.assertIn(b'Diff ticket 1', ret.data)
        self.assertIn(b'<a href="/#/ticket/1">Ticket 1</a>', ret.data)
        self.assertIn(b'type="submit" value="Regenerate Diffs">', ret.data)

    def test_all_logs(self):
        """Ensure the list of all logs is sensible."""
        self.test_get_log()  # Creates a dummy log to find.
        ret = self.app.get('/log/')
        self.assertEqual(ret.status_code, 200)
        self.assertIn(b'<a href="/log/1/build/1/info">', ret.data)
        self.assertIn(b'<span class="success_">', ret.data)
        open(self.tempdir + '/log/1/build/1/success_True', 'w').close()
        open(self.tempdir + '/log/1/build/1/stamp_just now', 'w').close()
        ret = self.app.get('/log/1/')
        self.assertEqual(ret.status_code, 200)
        self.assertIn(b'<a href="/log/1/build/1/info">', ret.data)
        self.assertIn(b'<span class="success_True">', ret.data)
        self.assertIn(b'<td>just now</td>', ret.data)

    @patch('bileto.streams.walk')
    def test_log_history(self, walk):
        """Ensure /v1/tickets gives empty result set."""
        walk.return_value = [
            ('foo/1', None, ('info', 'debug', 'stamp_foo', 'success_True')),
            ('bar/2', None, ('info', 'debug', 'stamp_bar', 'success_True')),
            ('baz/3', None, ('info', 'debug', 'stamp_baz', 'success_False')),
        ]
        ret = self.app.get('/log/1/build/')
        self.assertEqual(ret.status_code, 200)
        self.assertIn(
            b'class="success_False">&nbsp;</span>\n<a href="3">\n<strong>'
            b'3.</strong>\n<span class="stamp">baz</span>\n</a>\n', ret.data)
        self.assertIn(
            b'<a href="2">\n<strong>2.</strong>\n<span class="stamp">bar',
            ret.data)
        self.assertIn(
            b'class="success_True">&nbsp;</span>\n<a href="1">\n<strong>1.'
            b'</strong>\n<span class="stamp">foo',
            ret.data)
        self.assertIn(
            b'<input type="checkbox" name="includes" value="qtmir">qtmir</',
            ret.data)
        self.assertNotIn(b'qtmir-gles', ret.data)
        self.assertIn('qtmir-gles', Request.query.get(1).sources_set)
        walk.assert_called_once_with(self.tempdir + '/log/1/build')

    def test_bad_auth(self):
        """Reject users with bad permissions."""
        with self.app.session_transaction() as sesh:
            sesh['teams'] = ['ci-train-nonusers']
            sesh['nickname'] = 'mal'
        ret = self.app.post('/log/1/build/latest/')
        self.assertEqual(ret.status_code, 401)

    def test_bad_level(self):
        """Reject requests with bad debug levels."""
        ret = self.app.get('/log/1/build/1/wonka/')
        self.assertEqual(ret.status_code, 400)

    def test_get_log(self, message=None):
        """Fetch fresh log files."""
        base = '/log/1/build/1'
        path = base + '/info/'
        local = self.tempdir + base
        makedirs(local, exist_ok=True)
        with open(local + '/info', 'w', encoding='utf-8') as data:
            data.write(message or 'hi from tests')
        ret = self.app.get(path)
        self.assertEqual(ret.status_code, 200)
        if message is None:
            self.assertIn(b'hi from tests', ret.data)
        return ret

    def test_get_log_still_running(self):
        """Let users know when a log is incomplete."""
        path = join(self.tempdir, 'log/1/build/1/')
        makedirs(path, exist_ok=True)
        with open(path + 'info', 'w') as locked:
            flock(locked, LOCK_EX)
            ret = self.test_get_log('still running')
            self.assertIn(b'Page will automatically reload every 2', ret.data)

    def test_urls(self):
        """Make sure URLs are URLified."""
        ret = self.test_get_log('visit http://example.com/fresh')
        self.assertIn(
            b'visit <a href="http://example.com/fresh">fresh</a>',
            ret.data)

    def test_urls_2(self):
        """Handle URL corner cases."""
        ret = self.test_get_log('<http://example.com/ugh>')
        self.assertIn(
            b'&lt;<a href="http://example.com/ugh">ugh</a>&gt;',
            ret.data)

    def test_404_log(self):
        """404 when no log found."""
        ret = self.app.get('/log/500/build/1000/')
        self.assertEqual(ret.status_code, 404)

    @patch('bileto.streams.Popen')
    def test_run_subprocess(self, popen):
        """Stream output from subprocess."""
        ret = self.app.post('/log/1/build/latest/')
        self.assertIn(b'Redirecting...', ret.data)
        self.assertEqual(ret.status_code, 303)
        env = popen.mock_calls[0][2]['env']
        self.assertSequenceEqual(popen.mock_calls, [
            call((join(Config.ROOT, 'bileto', 'worker', 'main.py'),), env=env),
        ])
        self.assertEqual(env['ACTION'], 'build')
        self.assertEqual(env['REQUEST_PATH'], '/log/1/build/latest/')
