"""Bileto V1 Tests

Test all aspects of the JSON API, v1.
"""

import re

from json import dumps, loads
from mock import patch

from bileto.models import Request, Comment
from bileto.app import db

from tests.tests import BiletoTestCase


FAKE_LOG = '/tmp/bileto.test.log'
DATE = re.compile(r'^\d\d\d\d-\d\d-\d\dT\d\d:\d\d:\d\d')


class FakeConfig:
    """Dummy data to replace Config object."""
    api_token = 'faken'
    instance_id = 'testing-instance'
    jenkins_url = 'https://ci-train.staging.ubuntu.com/'
    ppa_team = 'team-not-set'


def parse(blob):
    """Simple wrapper for loads()."""
    return loads(blob.decode('utf-8'))


@patch('bileto.v1.Config', FakeConfig())
@patch('bileto.auth.Config', FakeConfig())
class V1TestCase(BiletoTestCase):
    """Test JSON API v1."""

    def setUp(self):
        """Fake login to easier test authenticated endpoints."""
        super().setUp()
        with self.app.session_transaction() as sesh:
            sesh['teams'] = ['ubuntu-core-dev']
            sesh['nickname'] = 'fritz'

    def test_dashboard_redirect(self):
        """Ensure dashboard redirects to front page."""
        ret = self.app.get('/static/dashboard.html', follow_redirects=False)
        self.assertIn(b'<a href="/">/</a>.', ret.data)
        self.assertEqual(ret.status_code, 302)

    def test_get_tickets(self):
        """Ensure /v1/tickets gives empty result set."""
        ret = self.app.get('/v1/tickets')
        data = parse(ret.data)
        self.assertEqual(data['nickname'], 'fritz')
        self.assertEqual(data['assigned'], 0)
        self.assertEqual(len(data['requests']), 0)
        self.assertEqual(ret.status_code, 200)

    def test_bad_authentication(self):
        """Ensure /v1/tickets can't be POSTed to without authentication."""
        with self.app.session_transaction() as sesh:
            sesh['teams'] = []
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(request_id=1, description='bar')),
            content_type='application/json')
        self.assertIn(b'401 Unauthorized', ret.data)
        self.assertEqual(ret.status_code, 401)

    def test_missing_json(self):
        """Ensure POSTing to /v1/tickets requires JSON."""
        ret = self.app.post('/v1/tickets', data='')
        self.assertEqual(ret.status_code, 400)

    def test_create_ticket(self, desc='foo', silo='siloname', **kwargs):
        """Ensure a ticket can be created."""
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(description=desc, siloname=silo, **kwargs)),
            content_type='application/json')
        data = parse(ret.data)
        self.assertIn('request_id', data)
        self.assertEqual(ret.status_code, 201)
        return data['request_id']

    def test_create_ticket_api_token(self, desc='foo'):
        """Ensure you can create tickets with the API token also."""
        with self.app.session_transaction() as sesh:
            sesh['teams'] = []
        ret = self.app.post(
            '/v1/tickets?token=faken',
            data=dumps(dict(description=desc)),
            content_type='application/json')
        self.assertEqual(ret.status_code, 201)

    def test_create_ticket_null_api_token(self, desc='foo'):
        """Ensure invalid API tokens are rejected."""
        with self.app.session_transaction() as sesh:
            sesh['teams'] = []
        ret = self.app.post(
            '/v1/tickets?token=wrong',
            data=dumps(dict(description=desc)),
            content_type='application/json')
        self.assertIn(b'401 Unauthorized', ret.data)
        self.assertEqual(ret.status_code, 401)

    def test_update_ticket(self):
        """Ensure existing tickets can be updated."""
        self.test_create_ticket('foo')
        self.assertEqual(Request.query.get(1).description, 'foo')
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(
                request_id=1,
                status='Pub',
                job_log='/',
                published_versions='hi',
                siloname="landing-987",
                description='z')),
            content_type='application/json')
        self.assertEqual(ret.status_code, 200)
        req = Request.query.get(1)
        self.assertEqual(req.description, 'z')
        self.assertEqual(
            req.search,
            '\n\n\nfritz\nz\n\n\n\n/\n\n\n\nhi\nN/A\n\nlanding-987\n\nPub\n\n')
        comment = Comment.query.get(1)
        self.assertEqual(comment.text, 'Pub')
        self.assertEqual(comment.log_url, '/')
        self.assertEqual(comment.published_versions, 'hi')
        self.assertEqual(comment.siloname, 'landing-987')

    def test_update_ticket_no_dupe_comments(self):
        """Audit log doesn't get duplicate comments."""
        self.test_create_ticket('foo')
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(
                request_id=1,
                status='Currently building',
                job_log='/#/ticket/1')),
            content_type='application/json')
        self.assertEqual(ret.status_code, 200)
        comment = Comment.query.get(1)
        self.assertEqual(comment.text, 'Currently building')
        self.assertEqual(comment.log_url, '/#/ticket/1')
        self.assertIsNone(Comment.query.get(2))
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(request_id=1, status='Currently building')),
            content_type='application/json')
        self.assertEqual(ret.status_code, 200)
        self.assertIsNone(Comment.query.get(2))

    def test_update_clear_signoffs(self, message='Preparing packages'):
        """Clear signoffs when silo is rebuilt."""
        request_id = self.test_create_ticket(
            'foo',
            status='Successfully built',
            series='xenial+vivid',
            lander_signoff='Approved',
            britney_signoff='Approved',
            qa_signoff='Approved')
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(
                request_id=request_id,
                status='Publish failed',
                job_log='/#/ticket/{}'.format(request_id))),
            content_type='application/json')
        self.assertEqual(ret.status_code, 200)
        req = Request.query.get(request_id)
        self.assertEqual(req.lander_signoff, 'Approved')
        self.assertEqual(req.britney_signoff, 'Approved')
        self.assertEqual(req.qa_signoff, 'Approved')
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(
                request_id=request_id,
                status=message,
                job_log='/#/ticket/{}'.format(request_id))),
            content_type='application/json')
        self.assertEqual(ret.status_code, 200)
        req = Request.query.get(request_id)
        self.assertEqual(req.lander_signoff, '')
        self.assertEqual(req.britney_signoff, '')
        self.assertEqual(req.qa_signoff, 'Required')

    def test_udpate_clear_signoffs_moar(self):
        """Repeat the previous test with different messages."""
        self.test_update_clear_signoffs('Failed to build')
        self.test_update_clear_signoffs('Currently building')

    def test_update_ignore_invalid(self):
        """Ignore invalid keys."""
        self.test_create_ticket('foo')
        self.assertEqual(Request.query.get(1).description, 'foo')
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(request_id=1, description='bar', huh='wat')),
            content_type='application/json')
        self.assertEqual(ret.status_code, 200)
        req = Request.query.get(1)
        self.assertEqual(req.description, 'bar')

    def test_update_ticket_set_qa(self):
        """Ensure QA field is auto-set properly."""
        self.test_create_ticket('foo', lander_signoff='Approved')
        self.assertEqual(Request.query.get(1).description, 'foo')
        self.assertEqual(Request.query.get(1).qa_signoff, 'N/A')
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(request_id=1, series='xenial+vivid')),
            content_type='application/json')
        self.assertEqual(ret.status_code, 200)
        self.assertEqual(Request.query.get(1).qa_signoff, 'Required')
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(request_id=1, britney_signoff='Approved')),
            content_type='application/json')
        self.assertEqual(ret.status_code, 200)
        self.assertEqual(Request.query.get(1).qa_signoff, 'Ready')
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(
                request_id=1,
                download_links='foo',
                britney_signoff='',
                qa_signoff='Ready')),
            content_type='application/json')
        self.assertEqual(ret.status_code, 200)
        self.assertEqual(Request.query.get(1).qa_signoff, 'Ready')

    def test_update_ticket_set_britney(self):
        """Ensure britney signoff field is auto-set properly."""
        self.test_create_ticket('foo', lander_signoff='Approved')
        self.assertEqual(Request.query.get(1).britney_signoff, '')
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(
                request_id=1,
                siloname='ubu/100',
                status='Successfully built.',
                job_log='/',
            )),
            content_type='application/json')
        self.assertEqual(ret.status_code, 200)
        self.assertEqual(Request.query.get(1).britney_signoff, 'Queued')
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(request_id=1, britney_signoff='Running')),
            content_type='application/json')
        self.assertEqual(ret.status_code, 200)
        self.assertEqual(Request.query.get(1).britney_signoff, 'Running')

    def test_update_ticket_doesnt_exist(self):
        """Ensure Bileto doesn't explode when updating a nonexistent ticket."""
        self.test_create_ticket('foo')
        self.assertEqual(Request.query.get(1).description, 'foo')
        ret = self.app.post(
            '/v1/tickets',
            data=dumps(dict(request_id=100, description='bar')),
            content_type='application/json')
        self.assertEqual(ret.status_code, 400)
        req = Request.query.get(1)
        self.assertEqual(req.description, 'foo')
        self.assertEqual(req.creator, 'fritz')

    def test_date_format(self):
        """Ensure tickets have the correct date format."""
        self.test_create_ticket('timestamp')
        ret = self.app.get('/v1/tickets')
        data = parse(ret.data)
        assert DATE.match(data['requests'][-1]['date'])

    def test_search(self):
        """Ensure basic searching functionality of /v1/tickets works."""
        self.test_create_ticket('alpha')
        self.test_create_ticket('beta')
        self.test_create_ticket('delta')
        self.test_create_ticket('gamma')
        self.test_create_ticket('zeta')
        self.test_create_ticket('eta')
        ret = self.app.get('/v1/tickets?search=eta')
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 3)
        self.assertEqual(ret.status_code, 200)
        descs = [d.get('description') for d in data['requests']]
        self.assertIn('eta', descs)
        self.assertIn('zeta', descs)
        self.assertIn('beta', descs)
        self.assertNotIn('alpha', descs)
        self.assertNotIn('delta', descs)
        self.assertNotIn('gamma', descs)

    def test_search_active_only(self):
        """Ensure searching can hide old requests."""
        self.test_create_ticket('alpha')
        self.test_create_ticket('beta')
        self.test_create_ticket('delta')
        self.test_create_ticket('gamma')
        self.test_create_ticket('zeta')
        self.test_create_ticket('eta')
        for i in range(1, 7):
            Request.query.get(i).status = 'Landed'
        db.session.commit()
        ret = self.app.get('/v1/tickets?search=eta&active_only')
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 0)
        self.assertEqual(ret.status_code, 200)
        descs = [d.get('description') for d in data['requests']]
        self.assertEqual(descs, [])

    def test_search_by_field(self):
        """Ensure you can search by a specific field."""
        self.test_create_ticket('alpha')
        self.test_create_ticket('beta')
        self.test_create_ticket('delta')
        self.test_create_ticket('gamma')
        self.test_create_ticket('zeta')
        self.test_create_ticket('eta')
        ret = self.app.get('/v1/tickets?description=eta')
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 3)
        self.assertEqual(ret.status_code, 200)
        descs = [d.get('description') for d in data['requests']]
        self.assertIn('eta', descs)
        self.assertIn('zeta', descs)
        self.assertIn('beta', descs)
        self.assertNotIn('alpha', descs)
        self.assertNotIn('delta', descs)
        self.assertNotIn('gamma', descs)

    def test_search_by_multiple_field(self):
        """Ensure you can search by multiple fields."""
        self.test_create_ticket('alpha', '123')
        self.test_create_ticket('alpha', '124')
        self.test_create_ticket('alpha', '456')
        self.test_create_ticket('beta', '123')
        self.test_create_ticket('beta', '124')
        ret = self.app.get('/v1/tickets?description=alpha&siloname=12')
        data = parse(ret.data)
        self.assertEqual(ret.status_code, 200)
        result = ['{description}-{siloname}'.format(**req)
                  for req in data['requests']]
        self.assertSequenceEqual(result, ['alpha-124', 'alpha-123'])

    def test_search_by_field_none(self):
        """Ensure a search with no results gives expected output."""
        self.test_create_ticket('alpha')
        self.test_create_ticket('beta')
        self.test_create_ticket('delta')
        self.test_create_ticket('gamma')
        self.test_create_ticket('zeta')
        self.test_create_ticket('eta')
        ret = self.app.get('/v1/tickets?landers=eta')
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 0)
        self.assertEqual(ret.status_code, 200)
        descs = [d.get('description') for d in data['requests']]
        self.assertEqual(descs, [])

    def test_search_invalid_field(self):
        """Ensure that specifying an invalid field doesn't explode."""
        self.test_create_ticket('alpha')
        self.test_create_ticket('beta')
        ret = self.app.get('/v1/tickets?invalid=field')
        self.assertEqual(ret.status_code, 400)

    def test_pagination(self):
        """Ensure pagination behaves sanely."""
        self.test_create_ticket('alpha')
        self.test_create_ticket('beta')
        self.test_create_ticket('gamma')
        self.test_create_ticket('delta')
        ret = self.app.get('/v1/tickets?page=1&limit=2')
        self.assertEqual(ret.status_code, 200)
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 2)
        self.assertIn('page=2', data['next'])
        self.assertEqual(data['prev'], '')
        self.assertEqual(data['requests'][0]['description'], 'delta')
        self.assertEqual(data['requests'][1]['description'], 'gamma')
        ret = self.app.get('/v1/tickets?page=2&limit=2')
        self.assertEqual(ret.status_code, 200)
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 2)
        self.assertIn('page=1', data['prev'])
        self.assertEqual(data['next'], '')
        self.assertEqual(data['requests'][0]['description'], 'beta')
        self.assertEqual(data['requests'][1]['description'], 'alpha')

    def test_invalid_pagination(self):
        """Ensure pagination won't explode with invalid inputs."""
        self.test_create_ticket('alpha')
        self.test_create_ticket('beta')
        self.test_create_ticket('gamma')
        self.test_create_ticket('delta')
        ret = self.app.get('/v1/tickets?page=0&limit=1')
        self.assertEqual(ret.status_code, 200)
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 1)
        ret = self.app.get('/v1/tickets?page=&limit=1')
        self.assertEqual(ret.status_code, 200)
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 1)
        ret = self.app.get('/v1/tickets?page=-100&limit=1')
        self.assertEqual(ret.status_code, 200)
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 1)
        ret = self.app.get('/v1/tickets?page=foo&limit=1')
        self.assertEqual(ret.status_code, 200)
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 1)
        ret = self.app.get('/v1/tickets?page=1&limit=-1')
        self.assertEqual(ret.status_code, 200)
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 4)

    def test_colors(self):
        """Ensure requests have the correct colors."""
        self.test_create_ticket(
            status='Successfully built.',
            lander_signoff='Approved',
            britney_signoff='Approved',
            qa_signoff='N/A')
        self.test_create_ticket(
            status='Successfully built',
            lander_signoff='Approved',
            britney_signoff='Approved',
            qa_signoff='Approved')
        self.test_create_ticket(
            status='anything',
            series='xenial+vivid',
            lander_signoff='Approved',
            qa_signoff='Failed')
        self.test_create_ticket(status='anything')
        self.test_create_ticket(status='caught the flu')
        self.test_create_ticket(status='exception raised')
        self.test_create_ticket(status='error: can\'t fail')
        self.test_create_ticket(status='merging')
        self.test_create_ticket(status='Ready to Rock')
        self.test_create_ticket(status='DIRTY DANCING')
        self.test_create_ticket(status='silo needs action')
        self.test_create_ticket(status='FAILED: at life')
        ret = self.app.get('/v1/tickets')
        self.assertEqual(ret.status_code, 200)
        reqs = parse(ret.data)['requests']
        colors = [req['status_color'] for req in reqs]
        # Remember these are newest-first.
        self.assertSequenceEqual(colors, [
            'red', 'olive', 'black', 'black', 'black', 'red', 'red',
            'red', 'black', 'red', 'blue', 'blue'])

    def test_trainguards(self):
        """Ensure trainguards endpoint only shows tickets needing action."""
        self.test_create_ticket(
            'alpha',
            dest='stable-phone-overlay',
            series='vivid',
            status='Successfully built',
            lander_signoff='Approved',
            britney_signoff='Approved',
            qa_signoff='Approved')
        self.test_create_ticket(
            'beta',
            status='Successfully built.',
            lander_signoff='Approved',
            britney_signoff='Approved',
            qa_signoff='N/A')
        self.test_create_ticket(
            'delta',
            series='xenial+vivid',
            status='b0rked',
            lander_signoff='Approved',
            britney_signoff='Approved',
            qa_signoff='Approved')
        self.test_create_ticket(
            'gamma',
            status='bad',
            lander_signoff='Approved',
            britney_signoff='Approved',
            qa_signoff='N/A')
        self.test_create_ticket(
            'zeta',
            series='xenial+vivid',
            status='Successfully built',
            qa_signoff='Failed')
        self.test_create_ticket(
            'eta',
            series='xenial+vivid',
            status='bad',
            qa_signoff='Failed')
        self.test_create_ticket(
            'epsilon',
            dest='stable-phone-overlay',
            series='vivid',
            status='fred not authorized to upload',
            lander_signoff='Approved',
            britney_signoff='Approved',
            qa_signoff='Approved')
        self.test_create_ticket(
            'iota',
            status='packaging changes need manual ACKing',
            lander_signoff='Approved',
            britney_signoff='Approved',
            qa_signoff='N/A')
        self.test_create_ticket(
            'kappa',
            status='packaging changes need manual ACKing',
            series='xenial+vivid',
            qa_signoff='')
        self.test_create_ticket(
            'lambda',
            status='Landed',
            series='xenial+vivid',
            lander_signoff='Approved',
            britney_signoff='Approved',
            qa_signoff='Approved')
        ret = self.app.get('/v1/publishable')
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 4)
        self.assertEqual(ret.status_code, 200)
        descs = [d.get('description') for d in data['requests']]
        self.assertEqual(descs, ['alpha', 'beta', 'epsilon', 'iota'])

    def test_failures(self):
        """Ensure failures endpoint only shows broken stuff."""
        self.test_create_ticket('alpha', status='Can\'t publish: wah')
        self.test_create_ticket('beta', status='Build FAILed')
        self.test_create_ticket('delta', status='dch error')
        self.test_create_ticket('gamma', status='error: can\'t fail')
        self.test_create_ticket('zeta', status='Successfully built')
        self.test_create_ticket('eta', status='bad')
        ret = self.app.get('/v1/failures')
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 4)
        self.assertEqual(ret.status_code, 200)
        descs = [d.get('description') for d in data['requests']]
        self.assertEqual(descs, ['alpha', 'beta', 'delta', 'gamma'])

    def test_by_siloname_unique(self):
        """Ensure active tickets can be searched by siloname."""
        self.test_create_ticket(silo='ubuntu/landing-001')
        self.test_create_ticket(silo='ubuntu/landing-010')
        self.test_create_ticket(silo='ubuntu/landing-021')
        self.test_create_ticket(silo='ubuntu/landing-022')
        Request.query.get(1).status = 'Landed'
        db.session.commit()
        ret = self.app.get('/v1/silo/1')
        self.assertEqual(ret.status_code, 200)
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 2)
        self.assertEqual(data['requests'][0]['siloname'], 'ubuntu/landing-021')
        self.assertEqual(data['requests'][1]['siloname'], 'ubuntu/landing-010')

    def test_by_siloname_nonunique(self):
        """Ensure active tickets can be searched by siloname."""
        self.test_create_ticket(silo='ubuntu/landing-001')
        self.test_create_ticket(silo='ubuntu/landing-010')
        self.test_create_ticket(silo='ubuntu/landing-021')
        self.test_create_ticket(silo='ubuntu/landing-022')
        Request.query.get(1).status = 'Landed'
        db.session.commit()
        ret = self.app.get('/v1/silo/ubuntu/landing-010')
        self.assertEqual(ret.status_code, 200)
        data = parse(ret.data)
        self.assertEqual(len(data['requests']), 1)
        self.assertEqual(data['requests'][0]['siloname'], 'ubuntu/landing-010')

    def test_by_user(self):
        """Ensure active tickets can be searched by user."""
        self.test_create_ticket(landers='fritz a')
        self.test_create_ticket(landers='fritz b')
        self.test_create_ticket(landers='somebody else')
        self.test_create_ticket(landers='also not that first guy')
        Request.query.get(1).status = 'Landed'
        db.session.commit()
        ret = self.app.get('/v1/user/fritz')
        self.assertEqual(ret.status_code, 200)
        data = parse(ret.data)
        self.assertEqual(data['nickname'], 'fritz')
        self.assertEqual(len(data['requests']), 1)
        self.assertEqual(data['requests'][0]['landers'], 'fritz b')

    def test_get_one_ticket(self):
        """Ensure tickets can be retrieved by id."""
        self.test_create_ticket('alpha', silo='ubuntu/landing-001')
        ret = self.app.get('/v1/ticket/1')
        data = parse(ret.data)
        req = data['requests'][0]
        self.assertEqual(data['assigned'], 1)
        self.assertEqual(ret.status_code, 200)
        self.assertEqual(req['description'], 'alpha')
        self.assertEqual(req['request_id'], 1)
        self.assertEqual(req['status'], 'New')

    def test_get_one_ticket_404(self):
        """Invalid ticket numbers should 404."""
        self.test_create_ticket('alpha', silo='ubuntu/landing-001')
        ret = self.app.get('/v1/ticket/999')
        self.assertEqual(ret.status_code, 404)

    def test_create(self):
        """Ensure the create endpoint sends out an empty dummy."""
        ret = self.app.get('/v1/create')
        data = parse(ret.data)['requests']
        self.assertEqual(ret.status_code, 200)
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['distribution'], 'ubuntu')

    @patch('bileto.v1.lp')
    def test_metadata(self, lp_mock):
        """Ensure we get sensible metadata."""
        lp_mock.active_series.return_value = ['alpha', 'beta']
        ret = self.app.get('/v1/metadata')
        self.assertDictEqual(parse(ret.data), dict(
            docs=dict(
                artifacts='Artifacts',
                autopkgtest='Automated Test Results\n\nAutomatically filled '
                            'with links to britney excuses.html including '
                            'autopkgtest results.',
                britney_signoff='Automated Signoff\n\nIndicates if '
                                'britney has approved the silo based on '
                                'autopkgtests & other factors. Check the '
                                'excuses page for details.',
                comments='Audit Log',
                creator='Request Creator\n\nLaunchpad nickname of the person '
                        'who created this request.',
                date='Creation Date',
                description='Landing Description\n\nBriefly describe what bug '
                            'you\'re fixing or feature you\'re adding.',
                dest='Destination\n\nWhether silo targets ubuntu archive or '
                     'overlay PPA. No effect on dual silos.',
                distribution='Target Distribution\n\nWhat distribution to '
                             'build for, eg, ubuntu or ubuntu-rtm.',
                download_links='Manual Download URLs\n\nPlace links to click '
                               'packages and tarballs here.',
                job_log='URL of full status log\n\nThis is the full log '
                        'leading to the creation of the current status.',
                lander_signoff='Lander Signoff\n\nIndicates whether or '
                               'not the silo has been tested by the upstream '
                               'lander.',
                landers='Lander IRC Nicks\n\nSpace separated list of IRC nicks'
                        ' of the people responsible for the silo.',
                merge_proposals='Merge Proposal URLs\n\nSpace or newline '
                                'separated list of MERGE URLs. No branches!',
                published_versions='Deprecated.',
                qa_signoff='QA Signoff\n\nIndicates whether or not the '
                           'silo has been approved by QA team.',
                request_id='Request ID',
                siloname='Name of Assigned Silo\n\nSet automatically by '
                         'Bileto, do not touch.',
                search='Searchable Field',
                series='Target Series\n\nWhat series to build for, eg, wily, '
                       'vivid, or yakkety+xenial+vivid.',
                sources='Source Package Names\n\nAutomatically-discovered '
                        'list of source packages found in PPA.',
                status='Status\n\nSet automatically by Bileto, do not touch.',
                status_color='Status Color',
                sync_request='Sync Source\n\nWhere should this silo be synced '
                             'from? Leave blank if not syncing.\nValid values '
                             'are \'[silo number]\', \'[distro],[series]\', '
                             '\'[ppa],[series]\'\nor \'stable-overlay,[series]'
                             '\'.',
                test_plan='Test Plan\n\nLink to a document that explains the '
                          'steps to verify this silo.',
            ),
            datalists=dict(
                dest=[
                    '',
                    'team-not-set/ubuntu/stable-phone-overlay',
                ],
                distribution=['ubuntu'],
                lander_signoff=['', 'Approved', 'Failed'],
                qa_signoff=['N/A', 'Required', 'Ready', 'Approved', 'Failed'],
                series=['yakkety+xenial+vivid', 'alpha', 'beta'],
                status=['New', 'Landed', 'Abandoned'],
                sync_request=[
                    'ubuntu,alpha',
                    'ubuntu,beta',
                    '0',
                    'stable-overlay,vivid',
                    'ubuntu-rtm,14.09',
                    'ppa:~team/ubuntu/ppa,vivid',
                ]),
            jenkins='https://ci-train.staging.ubuntu.com/',
            ppa_team='team-not-set',
        ))

    def test_post_comment(self):
        """Ensure we can post comments."""
        self.test_create_ticket('alpha')
        ret = self.app.post(
            '/v1/comment',
            data=dumps(dict(request_id=1, text='all systems go')),
            content_type='application/json')
        self.assertEqual(ret.status_code, 201)
        self.assertEqual(Comment.query.get(1).text, 'all systems go')
        ret = self.app.get('/v1/tickets')
        data = parse(ret.data)['requests'][0]['comments'][0]
        self.assertEqual(data['author'], 'fritz')
        self.assertEqual(data['id'], 1)
        self.assertEqual(data['request_id'], 1)
        self.assertEqual(data['text'], 'all systems go')

    def test_post_comment_by_bot(self):
        """Ensure we can inject comments with API token."""
        self.test_create_ticket('alpha')
        with self.app.session_transaction() as sesh:
            sesh.update(teams=[], nickname='')
        ret = self.app.post(
            '/v1/comment?token=faken',
            data=dumps(dict(
                request_id=1, author='job starter', text='all systems go')),
            content_type='application/json')
        self.assertEqual(ret.status_code, 201)
        self.assertEqual(Comment.query.get(1).text, 'all systems go')
        ret = self.app.get('/v1/tickets')
        data = parse(ret.data)['requests'][0]['comments'][0]
        self.assertEqual(data['author'], 'job starter')
        self.assertEqual(data['id'], 1)
        self.assertEqual(data['request_id'], 1)
        self.assertEqual(data['text'], 'all systems go')

    def test_post_comment_by_bot_update(self):
        """Ensure comments are generated by changing fields."""
        self.test_create_ticket('alpha')
        with self.app.session_transaction() as sesh:
            sesh.update(teams=[], nickname='')
        ret = self.app.post(
            '/v1/tickets?token=faken',
            data=dumps(dict(request_id=1, britney_signoff='Failed')),
            content_type='application/json')
        self.assertEqual(ret.status_code, 200)
        ret = self.app.post(
            '/v1/tickets?token=faken',
            data=dumps(dict(request_id=1, job_log='/', status='hi')),
            content_type='application/json')
        self.assertEqual(ret.status_code, 200)
        self.assertEqual(Comment.query.get(1).text, 'britney_signoff: Failed')
        self.assertEqual(Comment.query.get(2).text, 'hi')
        ret = self.app.get('/v1/tickets')
        data = parse(ret.data)['requests'][0]['comments']
        self.assertEqual(len(data), 2)
        self.assertEqual(data[1]['author'], 'bileto-bot')
        self.assertEqual(data[1]['id'], 1)
        self.assertEqual(data[1]['request_id'], 1)
        self.assertEqual(data[1]['text'], 'britney_signoff: Failed')
        self.assertEqual(data[1]['log_url'], '/#/ticket/1')
        self.assertEqual(data[0]['author'], 'bileto-bot')
        self.assertEqual(data[0]['id'], 2)
        self.assertEqual(data[0]['request_id'], 1)
        self.assertEqual(data[0]['text'], 'hi')
        self.assertEqual(data[0]['log_url'], '/')

    def test_post_comment_empty(self):
        """Don't explode when given empty comment."""
        self.test_create_ticket('alpha')
        ret = self.app.post(
            '/v1/comment',
            data=dumps(dict(request_id=1)),
            content_type='application/json')
        self.assertEqual(ret.status_code, 400)
        ret = self.app.get('/v1/tickets')
        data = parse(ret.data)['requests'][0]['comments']
        self.assertEqual(data, [])

    def test_comment_search(self):
        """Ensure comments are searchable."""
        self.test_create_ticket('one')
        db.session.add(Comment(request_id=1, author='me', text='hi'))
        db.session.add(Comment(request_id=1, author='me', text='there'))
        db.session.add(Comment(request_id=1, author='me', text='friend'))
        db.session.commit()
        data = parse(self.app.get('/v1/comments?text=e').data)['comments']
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0]['author'], 'me')
        self.assertEqual(data[0]['id'], 3)
        self.assertEqual(data[0]['text'], 'friend')
        self.assertEqual(data[1]['author'], 'me')
        self.assertEqual(data[1]['id'], 2)
        self.assertEqual(data[1]['text'], 'there')

    def test_comment_invalid_search(self):
        """Return HTTP 400 when invalid search parameters set."""
        ret = self.app.get('/v1/comments?beep=boop')
        self.assertEqual(ret.status_code, 400)

    @patch('bileto.v1.LOGFILE', FAKE_LOG)
    def test_errors(self):
        """Ensure we can read the log file."""
        with open(FAKE_LOG, 'w', encoding='utf-8') as log:
            log.write('hi')
        ret = self.app.get('/v1/errors')
        self.assertEqual(ret.status_code, 200)
        self.assertEqual(ret.data, b'hi')

    @patch('bileto.v1.LOGFILE', FAKE_LOG)
    def test_errors_unauthenticated(self):
        """Ensure unauthenticated people can't access the log."""
        with self.app.session_transaction() as sesh:
            sesh['teams'] = []
        with open(FAKE_LOG, 'w', encoding='utf-8') as log:
            log.write('hi')
        ret = self.app.get('/v1/errors')
        self.assertEqual(ret.status_code, 401)

    def test_comments_only_for_one(self):
        """Don't waste bandwidth by sending comments that won't be shown."""
        self.test_create_ticket('one')
        self.test_create_ticket('two')
        for i in range(10):
            db.session.add(Comment(request_id=1, author='me', text=str(i)))
        db.session.commit()
        ret = self.app.get('/v1/tickets')
        reqs = parse(ret.data)['requests']
        assert reqs
        for req in reqs:
            self.assertNotIn('comments', req)
            self.assertNotIn('merge_proposals', req)
            self.assertNotIn('search', req)
        ret = self.app.get('/v1/ticket/1')
        reqs = parse(ret.data)['requests']
        assert reqs
        for req in reqs:
            self.assertEqual(len(req['comments']), 10)
            self.assertEqual(req['comments'][0]['text'], '9')
            self.assertNotIn('search', req)
            self.assertEqual(req['merge_proposals'], '')
