"""Bileto Worker Main Test

Test the main subprocess script.
"""

from os import getcwd
from unittest.mock import call, patch

from tests.tests import WorkerTestCase

from bileto.models import BiletoError, Request
from bileto.worker.main import main


FAKE_ENV = dict(
    BLAME='robru',
    BLAME_EMAIL='robru@rob.ru',
    BLAME_FULLNAME='Test Runner Guy',
    REQUESTID='1',
    ACTION='build',
)


class MainTestCase(WorkerTestCase):
    """Test Main Subprocess Script."""

    @patch('bileto.worker.main.chdir')
    @patch('bileto.worker.main.rmtree')
    @patch('bileto.worker.main.mkdtemp')
    @patch('bileto.worker.main.Request')
    @patch('bileto.worker.main.Manager')
    @patch('bileto.worker.main.environ', FAKE_ENV)
    def test_main(self, Manager, reqmock, mkdtemp, rmtree, chdir):
        """Check main is doing some sensible things."""
        self.assertEqual(main(), 0)
        chdir.assert_called_once_with(mkdtemp.return_value)
        mkdtemp.assert_called_once_with('-bileto-workspace')
        self.assertSequenceEqual(Manager.mock_calls, [
            call(reqmock.query.get.return_value),
            call().validate(),
            call().validate().collect(),
            call().validate().collect().build(),
            call().validate().collect().build().upload(),
            call().validate().collect().build().upload().expose(),
            call().validate().collect().build().upload().expose().diff(),
        ])
        rmtree.assert_called_once_with(getcwd(), ignore_errors=True)
        reqmock.query.get.return_value.set_status.assert_called_once_with(
            'Preparing packages.')

    @patch('bileto.worker.main.chdir')
    @patch('bileto.worker.main.rmtree')
    @patch('bileto.worker.main.mkdtemp')
    @patch('bileto.worker.main.Request')
    @patch('bileto.worker.main.Manager')
    @patch('bileto.worker.main.environ', FAKE_ENV)
    def test_main_fail1(self, Manager, reqmock, mkdtemp, rmtree, chdir):
        """Fail safely when ticket not found."""
        reqmock.query.get.return_value = None
        self.assertEqual(main(), 1)
        chdir.assert_called_once_with(mkdtemp.return_value)
        mkdtemp.assert_called_once_with('-bileto-workspace')
        rmtree.assert_called_once_with(getcwd(), ignore_errors=True)

    @patch('bileto.worker.main.chdir')
    @patch('bileto.worker.main.rmtree')
    @patch('bileto.worker.main.mkdtemp')
    @patch('bileto.worker.main.Request')
    @patch('bileto.worker.main.Manager')
    @patch('bileto.worker.main.environ', FAKE_ENV)
    def test_main_fail2(self, Manager, reqmock, mkdtemp, rmtree, chdir):
        """Fail safely when BiletoError raised."""
        Manager.side_effect = BiletoError('zoinks!')
        self.assertEqual(main(), 2)
        self.assertEqual(
            reqmock.query.get.return_value.set_status.mock_calls,
            [call('Preparing packages.'), call('zoinks!')])
        chdir.assert_called_once_with(mkdtemp.return_value)
        mkdtemp.assert_called_once_with('-bileto-workspace')
        rmtree.assert_called_once_with(getcwd(), ignore_errors=True)

    @patch('bileto.worker.main.chdir')
    @patch('bileto.worker.main.rmtree')
    @patch('bileto.worker.main.mkdtemp')
    @patch('bileto.worker.main.Manager')
    @patch('bileto.worker.main.Request.query')
    @patch('bileto.worker.main.Request.set_status')
    @patch('bileto.worker.main.environ', FAKE_ENV)
    def test_main_lock(self, status, query, Manager, mkdtemp, rmtree, chdir):
        """Ensure lock prevents duplicate jobs running."""
        query.get.side_effect = lambda x: Request(
            request_id=1, series='jiminy+cricket')
        self.assertEqual(main(), 0)
        self.assertEqual(main(), 3)

    @patch('bileto.worker.main.chdir')
    @patch('bileto.worker.main.rmtree')
    @patch('bileto.worker.main.mkdtemp')
    @patch('bileto.worker.main.Request')
    @patch('bileto.worker.main.Manager')
    @patch('bileto.worker.main.environ', FAKE_ENV)
    def test_main_fail3(self, Manager, reqmock, mkdtemp, rmtree, chdir):
        """Fail safely when other exception raised."""
        Manager.side_effect = KeyError('user not found')
        self.assertEqual(main(), 4)
        chdir.assert_called_once_with(mkdtemp.return_value)
        mkdtemp.assert_called_once_with('-bileto-workspace')
        rmtree.assert_called_once_with(getcwd(), ignore_errors=True)
