"""Bileto Manager Test

Test the package manager class.
"""

from asyncio import coroutine
from collections import defaultdict
from unittest.mock import Mock, patch, call

from lazr.restfulclient.errors import BadRequest

from tests.tests import WorkerTestCase, fake_subproc

from bileto.lplib import lp
from bileto.models import BiletoError
from bileto.worker.manager import Manager, Merge, Manual, Package


DIFF_ENV = dict(INCLUDES='')
MAKE_PPA_ENV = dict(REQUEST_URL_ROOT='https://ppa/thing')
MERGES = fake_subproc(0, [(b'ofono', b'')] * 5)
MOCKS = defaultdict(Mock)
SLASH = '/'
N = '\n'


def mp(num):
    """Name an MP."""
    return '{}mp{}'.format(lp.MP_ROOT, num)


def mp_obj(num, prereq=None, target=None):
    """Return an MP mock."""
    num = str(num)[-1]
    mock = MOCKS[num]
    mock.source_branch_link = mock.web_link = mp(num)
    mock.target_branch.display_name = target or 'lp:ofono'
    if prereq:
        mock.prerequisite_branch_link = mp(int(num) + 1)
    return mock


class ManagerTestCase(WorkerTestCase):
    """Test Manager class."""

    @patch('bileto.worker.merge.lp.load')
    @patch('bileto.worker.vcs.create_subprocess_exec', MERGES)
    def test_manager_merges(self, load):
        """Ensure manager can instantiate merges."""
        load.side_effect = mp_obj
        ticket = Mock(merge_proposals_list=[mp(1), mp(2), mp(3)])
        manager = Manager(ticket)
        self.assertEqual(manager.merges(), dict(ofono=[
            mp_obj(1), mp_obj(2), mp_obj(3),
        ]))

    @patch('bileto.worker.merge.lp.load')
    @patch('bileto.worker.vcs.create_subprocess_exec', MERGES)
    def test_manager_merges_sorted(self, load):
        """Ensure manager can sort merges."""
        load.side_effect = lambda x: mp_obj(x, prereq=True)
        ticket = Mock(merge_proposals_list=[mp(1), mp(2), mp(3)])
        manager = Manager(ticket)
        self.assertEqual(manager.merges(), dict(ofono=[
            mp_obj(3), mp_obj(2), mp_obj(1),
        ]))

    @patch('bileto.worker.merge.lp.load')
    @patch('bileto.worker.vcs.create_subprocess_exec', MERGES)
    def test_manager_merges_invalid(self, load):
        """Halt when merges invalid."""
        load.return_value = Mock()
        ticket = Mock(merge_proposals_list=['http://blah'])
        manager = Manager(ticket)
        with self.assertRaisesRegex(BiletoError, 'Failed to understand'):
            manager.merges()

    @patch('bileto.worker.merge.lp.load')
    @patch('bileto.worker.vcs.create_subprocess_exec', MERGES)
    def test_manager_merges_invalid_targets(self, load):
        """Halt when not all merges target same trunk."""
        load.side_effect = lambda x: mp_obj(x, target=x)
        ticket = Mock(merge_proposals_list=[mp(1), mp(2), mp(3)])
        manager = Manager(ticket)
        with self.assertRaisesRegex(BiletoError, 'Too many merge targets'):
            manager.merges()

    @patch('bileto.worker.manager.environ', dict(INCLUDES='1 2,3,   4'))
    def test_manager_names_env(self):
        """Ensure manager can identify source package names from environ."""
        ticket = Mock()
        manager = Manager(ticket)
        self.assertEqual(manager.names(), {'1', '2', '3', '4'})

    @patch('bileto.worker.manager.Manager.ppa')
    @patch('bileto.worker.manager.Manager.merges')
    @patch('bileto.worker.manager.environ', dict(INCLUDES=''))
    def test_manager_names_ppa(self, merges, ppa):
        """Ensure manager can identify source package names from the PPA."""
        merges.return_value = []
        ppa.getPublishedSources.return_value = [
            Mock(source_package_name='foo', status='Published'),
            Mock(source_package_name='bar', status='Pending'),
            Mock(source_package_name='nope', status='Deleted'),
            Mock(source_package_name='nada', status='Superceded'),
        ]
        ticket = Mock(sources_set=set())
        manager = Manager(ticket)
        self.assertEqual(manager.names(), {'foo', 'bar'})
        self.assertEqual(ticket.sources, 'bar foo')

    @patch('bileto.worker.manager.Manager.names')
    @patch('bileto.worker.manager.Manager.merges')
    def test_manager_types(self, merges, names):
        """Correctly choose what types to build with."""
        merges.return_value = dict(foo=[Mock()])
        names.return_value = ['foo', 'foo-gles', 'bar', 'bar-gles']
        ticket = Mock()
        manager = Manager(ticket)
        self.assertEqual(manager.types(), {Merge: {'foo'}, Manual: {'bar'}})

    @patch('bileto.worker.manager.lp.load')
    @patch('bileto.worker.manager.Manager.types')
    @patch('bileto.worker.manager.Manager.get_ppa')
    @patch('bileto.worker.manager.Config.ppa_team', 'testers')
    def test_manager_packages(self, ppa, types, load):
        """Instantiate Package subclass objects."""
        def loader(name):
            """Instantiate a Mock with correct name."""
            ret = Mock()
            ret.name = name.split(SLASH)[-1]
            ret.web_link = 'https://' + name
            return ret
        load.side_effect = loader
        Package.instances.clear()
        types.return_value = {
            Merge: {'qtmir', 'webbrowser-app'},
            Manual: {'qtubuntu', 'oxide-qt'},
        }
        Merge.all_mps = {'qtmir': [Mock()], 'webbrowser-app': [Mock()]}
        ticket = Mock(
            distribution='ubuntu',
            series='yakkety+xenial+vivid',
            siloname='landing-123')
        manager = Manager(ticket)
        manager.dest.web_link = 'https://ubuntu'
        result = manager.packages()
        self.assertDictEqual({
            key: obj.__class__.__name__ for key, obj in result.items()
        }, {
            'yakkety/oxide-qt': 'Manual',
            'yakkety/qtmir': 'Merge',
            'yakkety/qtmir-gles': 'GLES',
            'yakkety/qtubuntu': 'Manual',
            'yakkety/qtubuntu-gles': 'Manual',
            'yakkety/webbrowser-app': 'Merge',
            'xenial/oxide-qt': 'Manual',
            'xenial/qtmir': 'Secondary',
            'xenial/qtmir-gles': 'GLES',
            'xenial/qtubuntu': 'Manual',
            'xenial/qtubuntu-gles': 'Manual',
            'xenial/webbrowser-app': 'Secondary',
            'vivid/oxide-qt': 'Manual',
            'vivid/qtmir': 'Secondary',
            'vivid/qtmir-gles': 'GLES',
            'vivid/qtubuntu': 'Manual',
            'vivid/qtubuntu-gles': 'Manual',
            'vivid/webbrowser-app': 'Secondary',
        })
        self.assertSequenceEqual(load.mock_calls, [
            call('ubuntu/yakkety'),
            call('ubuntu/xenial'),
            call('ubuntu/vivid'),
            call('~testers/+archive/ubuntu/stable-phone-overlay')
        ])

    @patch('bileto.worker.manager.lp.load')
    @patch('bileto.worker.manager.Manager.types')
    @patch('bileto.worker.manager.Manager.get_ppa')
    @patch('bileto.worker.manager.Config.ppa_team', 'testers')
    def test_manager_packages_none(self, ppa, types, load):
        """Raise a BiletoError if no packages are being considered."""
        Package.instances.clear()
        types.return_value = {
            Merge: set(),
            Manual: set(),
        }
        Merge.all_mps = {}
        ticket = Mock(
            distribution='ubuntu',
            series='yakkety+xenial+vivid',
            siloname='landing-123')
        manager = Manager(ticket)
        with self.assertRaisesRegex(BiletoError, 'No packages are being con'):
            manager.packages()

    @patch('bileto.worker.manager.Package')
    @patch('bileto.worker.manager.Manager.packages')
    def test_manager_passthrough(self, packages, package):
        """Ensure that arbitrary methods called on manager are passed down."""
        called = []

        @coroutine
        def appender():
            """Prove we were called by appending to list."""
            called.append(True)
        instance = Mock()
        instance.do_the_stuff = appender
        packages.return_value = dict(blip=instance, glorp=instance)
        ticket = Mock()
        manager = Manager(ticket)
        manager.do_the_stuff()
        package.pre_do_the_stuff.assert_called_once_with(ticket)
        package.post_do_the_stuff.assert_called_once_with(ticket)
        self.assertEqual(called, [True, True])

    @patch('bileto.worker.manager.Package')
    @patch('bileto.worker.manager.Manager.packages')
    @patch('bileto.worker.manager.environ', DIFF_ENV)
    def test_manager_diff(self, packages, package):
        """Force diffing against all packages even if not all were built."""
        called = []

        @coroutine
        def appender():
            """Prove we were called by appending to list."""
            called.append(True)
        instance = Mock()
        instance.diff = appender
        packages.return_value = dict(blip=instance, glorp=instance)
        ticket = Mock(sources='blip glorp')
        manager = Manager(ticket)
        manager.diff()
        package.pre_diff.assert_called_once_with(ticket)
        package.post_diff.assert_called_once_with(ticket)
        self.assertEqual(called, [True, True])
        self.assertEqual(DIFF_ENV, dict(INCLUDES='blip glorp'))
        packages.cache_clear.assert_called_once_with()

    @patch('bileto.worker.manager.lp')
    def test_manager_nuke(self, lp_mock):
        """Delete ephemeral PPA during nuke phase."""
        ticket = Mock(request_id='5')
        manager = Manager(ticket)
        manager.nuke()
        self.assertEqual(
            str(lp_mock.mock_calls),
            "[call.ppa_team.getPPAByName(name='05'),\n"
            " call.ppa_team.getPPAByName().web_link.__str__(),\n"
            " call.ppa_team.getPPAByName().lp_delete()]")

    @patch('bileto.worker.manager.lp')
    def test_manager_nuke_again(self, lp_mock):
        """Don't explode if the same PPA is nuked twice in a row."""
        lp_mock.ppa_team.getPPAByName.return_value.lp_delete.side_effect = (
            BadRequest(b'', b'Archive already deleted'))
        ticket = Mock(request_id='5')
        manager = Manager(ticket)
        manager.nuke()
        self.assertEqual(
            str(lp_mock.mock_calls),
            "[call.ppa_team.getPPAByName(name='05'),\n"
            " call.ppa_team.getPPAByName().web_link.__str__(),\n"
            " call.ppa_team.getPPAByName().lp_delete()]")

    @patch('bileto.worker.manager.lp')
    def test_manager_nuke_invalid(self, lp_mock):
        """Don't explode if a non-existing PPA is nuked."""
        lp_mock.ppa_team.getPPAByName.return_value.lp_delete.side_effect = (
            AttributeError)
        ticket = Mock(request_id='5')
        manager = Manager(ticket)
        manager.nuke()
        self.assertEqual(
            str(lp_mock.mock_calls),
            "[call.ppa_team.getPPAByName(name='05'),\n"
            " call.ppa_team.getPPAByName().web_link.__str__(),\n"
            " call.ppa_team.getPPAByName().lp_delete()]")

    @patch('bileto.worker.manager.lp')
    def test_manager_ppa_ephemeral(self, lp_mock):
        """Fetch an existing ephemeral PPA."""
        ticket = Mock(distribution='ubuntu', request_id='5000')
        manager = Manager(ticket)
        self.assertIsNotNone(manager.ppa_ephemeral)
        self.assertEqual(
            str(lp_mock.mock_calls),
            "[call.ppa_team.getPPAByName(name='5000'),\n"
            " call.ppa_team.getPPAByName().web_link.__str__(),\n"
            " call.ppa_team.getPPAByName().__bool__(),\n"
            " call.ppa_team.getPPAByName().name.__str__()]")

    @patch('bileto.worker.manager.lp')
    @patch('bileto.worker.manager.environ', MAKE_PPA_ENV)
    def test_manager_ppa_ephemeral_create(self, lp_mock):
        """Fetch an existing ephemeral PPA."""
        created = Mock()
        results = [created, None]
        lp_mock.ppa_team.getPPAByName.side_effect = lambda **_: results.pop()
        lp_mock.distributions = dict(testbuntu='worky')
        ticket = Mock(distribution='testbuntu', path='/me', request_id='6000')
        manager = Manager(ticket)
        self.assertIsNotNone(manager.ppa_ephemeral)
        self.assertEqual(
            str(lp_mock.mock_calls),
            "[call.ppa_team.getPPAByName(name='6000'),\n"
            " call.ppa_team.createPPA(description='https://ppa/thing/me', "
            "displayname='Bileto 6000', distribution='worky', name='6000'),\n"
            " call.ppa_team.getPPAByName(name='6000')]")

    def test_manager_ppa_unassigned(self):
        """Fetch an existing ephemeral PPA."""
        ticket = Mock(siloname='')
        manager = Manager(ticket)
        with self.assertRaisesRegex(BiletoError, 'must assign this'):
            assert manager.ppa

    def test_manager_dest_vivid(self):
        """Pin vivid silos to overlay always."""
        ticket = Mock(series='vivid')
        manager = Manager(ticket)
        self.assertEqual(manager.dest, manager.overlay)

    @patch('bileto.worker.manager.lp')
    def test_manager_dest_dual(self, lp_mock):
        """Pin duals to main archive always."""
        ticket = Mock(series='yakkety+xenial+vivid', dest='team/ubuntu/ppa')
        manager = Manager(ticket)
        self.assertEqual(manager.dest, lp_mock.ubuntu.main_archive)

    @patch('bileto.worker.manager.lp')
    def test_manager_dest_ppa(self, lp_mock):
        """Allow individual series to target PPAs."""
        ticket = Mock(series='yakkety', dest='team/ubuntu/ppa')
        manager = Manager(ticket)
        self.assertEqual(manager.dest, lp_mock.load.return_value)
        lp_mock.load.assert_called_once_with('~team/+archive/ubuntu/ppa')

    @patch('bileto.worker.manager.lp')
    def test_manager_dest_default(self, lp_mock):
        """No dest means main archive."""
        ticket = Mock(series='yakkety', dest='')
        manager = Manager(ticket)
        self.assertEqual(manager.dest, lp_mock.ubuntu.main_archive)
