"""Bileto Merge Test

Test that Merges behave correctly.
"""

from unittest.mock import Mock, call, patch
from subprocess import PIPE
from os.path import join

from tests.tests import WorkerTestCase, fake_subproc

from bileto.settings import Config
from bileto.worker.merge import Merge


STD = (b'', b'')
COLLECT = fake_subproc(0, [
    STD, STD, STD,
    (b'0.9', b''), (b'1.0', b''),
    (b'Jimmy Jo Bob <j@bob.com>, ', b''), STD,
])
MESSAGE = fake_subproc(0, [
    STD, STD, STD,
    (b'0.9', b''), (b'1.0', b''),
    STD, STD,
])
VERSIONS = fake_subproc(0, [
    STD, STD, STD,
    (b'1.0+16.04.20160412.1-0ubuntu1', b''),
    (b'1.0+16.04.20160412-0ubuntu1', b''),
    (b'Jimmy <j@bob.com>, Sally <s@lly.com>, Suzy <su@zy.com>, ', b''), STD,
])
EXPOSE = fake_subproc(0, [STD])


class MergeTestCase(WorkerTestCase):
    """Test Merge Packages."""

    def setUp(self):
        """Create some mocks."""
        super().setUp()
        bug1 = Mock()
        bug1.id = 123456
        bug2 = Mock()
        bug2.id = 654321
        self.mp = Mock()
        self.mp.commit_message = 'Fix bug #NNNNNN'
        self.mp.target_branch.display_name = 'lp:ofono'
        self.mp.target_branch.project.name = 'ofono'
        self.mp.source_branch.web_link = 'http://mp'
        self.mp.source_branch.linked_bugs = [bug1, bug2]

    @patch('bileto.worker.version.datetime')
    @patch('bileto.worker.vcs.create_subprocess_exec', COLLECT)
    def test_merge_collect(self, datetime):
        """Invoke merge collection correctly."""
        datetime.utcnow.return_value.strftime.return_value = '20160412'
        Merge.all_mps = dict(ofono=[self.mp])
        merge = Merge('ofono', self.distro, self.series, self.dest, self.ppa)
        self.ppa.getPublishedSources.return_value = [
            Mock(source_package_version='1.0+16.04.20160412-0ubuntu1')]
        self.run_coro(merge.collect())
        env = lambda x: COLLECT.mock_calls[x][2]['env']
        self.assertEqual(env(0)['SOURCE'], 'ofono')
        self.assertEqual(env(0)['SERIES'], 'xenial')
        self.assertEqual(env(0)['ACTION'], 'branch')
        self.assertEqual(env(1)['ACTION'], 'merge')
        self.assertEqual(env(1)['BUGS'], ' (LP: #123456, #654321)')
        self.assertEqual(env(2)['ACTION'], 'source_version')
        self.assertEqual(env(3)['ACTION'], 'source_version')
        self.assertEqual(env(4)['ACTION'], 'source_version')
        self.assertEqual(env(5)['ACTION'], 'append_changelog')
        self.assertEqual(env(5)['DEBFULLNAME'], 'Jimmy Jo Bob')
        self.assertEqual(env(5)['DEBEMAIL'], 'j@bob.com')
        self.assertEqual(env(5)['VERSION'], '1.0+16.04.20160412.1-0ubuntu1')
        self.assertEqual(env(6)['ACTION'], 'tag_release')
        script = join(Config.ROOT, 'scripts', 'vcs.sh')
        self.assertSequenceEqual(COLLECT.mock_calls, [
            call(script, 'lp:ofono', env=env(0), stdout=PIPE, stderr=PIPE),
            call(script, 'http://mp', env=env(1), stdout=PIPE, stderr=PIPE),
            call(script, 'lp:ofono', env=env(2), stdout=PIPE, stderr=PIPE),
            call(script, 'lp:~/ofono/ofono-ubuntu-xenial-12321',
                 env=env(3), stdout=PIPE, stderr=PIPE),
            call(script, 'WORKDIR', env=env(4), stdout=PIPE, stderr=PIPE),
            call(script, 'lp:ofono', env=env(5), stdout=PIPE, stderr=PIPE),
            call(script, 'lp:ofono', env=env(6), stdout=PIPE, stderr=PIPE),
        ])
        datetime.utcnow.return_value.strftime.assert_called_once_with('%Y%m%d')

    @patch('bileto.worker.version.datetime')
    @patch('bileto.worker.vcs.create_subprocess_exec', VERSIONS)
    def test_merge_collect_versions(self, datetime):
        """Invoke merge collection with correct versions."""
        datetime.utcnow.return_value.strftime.return_value = '20160412'
        Merge.all_mps = dict(ofono=[self.mp])
        merge = Merge('ofono', self.distro, self.series, self.dest, self.ppa)
        self.ppa.getPublishedSources.return_value = [
            Mock(source_package_version='1.0')]
        self.run_coro(merge.collect())
        env = lambda x: VERSIONS.mock_calls[x][2]['env']
        self.assertEqual(env(0)['SOURCE'], 'ofono')
        self.assertEqual(env(0)['SERIES'], 'xenial')
        self.assertEqual(env(0)['ACTION'], 'branch')
        self.assertEqual(env(1)['ACTION'], 'merge')
        self.assertEqual(env(1)['BUGS'], ' (LP: #123456, #654321)')
        self.assertEqual(env(2)['ACTION'], 'source_version')
        self.assertEqual(env(3)['ACTION'], 'source_version')
        self.assertEqual(env(4)['ACTION'], 'source_version')
        self.assertEqual(env(5)['VERSION'], '1.0+16.04.20160412.2-0ubuntu1')
        self.assertEqual(env(5)['DEBFULLNAME'], 'Jimmy, Sally, Suzy')
        self.assertEqual(env(5)['DEBEMAIL'], 'j@bob.com, s@lly.com, su@zy.com')
        self.assertEqual(env(5)['ACTION'], 'append_changelog')
        self.assertEqual(env(6)['ACTION'], 'tag_release')
        script = join(Config.ROOT, 'scripts', 'vcs.sh')
        self.assertSequenceEqual(VERSIONS.mock_calls, [
            call(script, 'lp:ofono', env=env(0), stdout=PIPE, stderr=PIPE),
            call(script, 'http://mp', env=env(1), stdout=PIPE, stderr=PIPE),
            call(script, 'lp:ofono', env=env(2), stdout=PIPE, stderr=PIPE),
            call(script, 'lp:~/ofono/ofono-ubuntu-xenial-12321',
                 env=env(3), stdout=PIPE, stderr=PIPE),
            call(script, 'WORKDIR', env=env(4), stdout=PIPE, stderr=PIPE),
            call(script, 'lp:ofono', env=env(5), stdout=PIPE, stderr=PIPE),
            call(script, 'lp:ofono', env=env(6), stdout=PIPE, stderr=PIPE),
        ])
        datetime.utcnow.return_value.strftime.assert_called_once_with('%Y%m%d')

    @patch('bileto.worker.version.datetime')
    @patch('bileto.worker.vcs.create_subprocess_exec', MESSAGE)
    def test_merge_collect_manual_changelog(self, datetime):
        """Allow manual debian/changelog entries."""
        self.mp.commit_message = None
        datetime.utcnow.return_value.strftime.return_value = '20160412'
        Merge.all_mps = dict(ofono=[self.mp])
        merge = Merge('ofono', self.distro, self.series, self.dest, self.ppa)
        self.ppa.getPublishedSources.return_value = []
        self.run_coro(merge.collect())
        datetime.utcnow.return_value.strftime.assert_called_once_with('%Y%m%d')

    @patch('bileto.worker.vcs.create_subprocess_exec', EXPOSE)
    def test_merge_expose(self):
        """Expose merges"""
        Merge.all_mps = dict(ofono=[self.mp])
        merge = Merge('ofono', self.distro, self.series, self.dest, self.ppa)
        self.run_coro(merge.expose())
        env = EXPOSE.mock_calls[0][2]['env']
        self.assertEqual(env['SOURCE'], 'ofono')
        self.assertEqual(env['SERIES'], 'xenial')
        self.assertEqual(env['ACTION'], 'expose')
        self.assertSequenceEqual(EXPOSE.mock_calls, [
            call(join(Config.ROOT, 'scripts', 'vcs.sh'),
                 'lp:~/ofono/ofono-ubuntu-xenial-12321',
                 env=env, stdout=PIPE, stderr=PIPE),
        ])
