"""Bileto Package Test

Test the steps for building a package.
"""

from os.path import join
from subprocess import PIPE
from unittest.mock import Mock, call, patch

from tests.tests import WorkerTestCase, fake_subproc

from bileto.settings import Config
from bileto.worker.package import Package


BUILD = fake_subproc(0, [(b'', b'')])
UPLOAD = fake_subproc(0, [(b'', b'')])
DIFF = fake_subproc(0, [(b'', b''), (b'', b'')])
POST_DIFF = fake_subproc(0, [(b'swift/diff (9000 lines)', b'')])


class PackageTestCase(WorkerTestCase):
    """Test Packages."""

    def test_package_post_validate(self):
        """Do global validation steps."""
        ticket = Mock()
        package = Package('foo', self.distro, self.series, self.dest, self.ppa)
        package.post_validate(ticket)
        self.assertEqual(ticket.mock_calls, [
            call.validate(),
            call.summarize(),
        ])

    @patch('bileto.worker.vcs.create_subprocess_exec', BUILD)
    def test_package_build(self):
        """Invoke package building."""
        package = Package('bar', self.distro, self.series, self.dest, self.ppa)
        self.run_coro(package.build())
        env = BUILD.mock_calls[0][2]['env']
        self.assertEqual(env['SOURCE'], 'bar')
        self.assertEqual(env['SERIES'], 'xenial')
        self.assertEqual(env['ACTION'], 'build')
        self.assertSequenceEqual(BUILD.mock_calls, [
            call(join(Config.ROOT, 'scripts', 'vcs.sh'),
                 env=env, stdout=PIPE, stderr=PIPE),
        ])

    @patch('bileto.worker.package.Config.ppa_team', 'uploaders')
    @patch('bileto.worker.vcs.create_subprocess_exec', UPLOAD)
    def test_package_upload(self):
        """Invoke package uploading."""
        package = Package('baz', self.distro, self.series, self.dest, self.ppa)
        self.run_coro(package.upload())
        env = UPLOAD.mock_calls[0][2]['env']
        self.assertEqual(env['SOURCE'], 'baz')
        self.assertEqual(env['SERIES'], 'xenial')
        self.assertEqual(env['DISTRO'], 'ubuntu')
        self.assertEqual(env['PPA_NAME'], '12321')
        self.assertEqual(env['PPA_TEAM'], 'uploaders')
        self.assertEqual(env['ACTION'], 'upload')
        self.assertSequenceEqual(UPLOAD.mock_calls, [
            call(join(Config.ROOT, 'scripts', 'vcs.sh'),
                 env=env, stdout=PIPE, stderr=PIPE),
        ])

    @patch('bileto.worker.vcs.create_subprocess_exec', DIFF)
    def test_package_diff(self):
        """Invoke package diffing."""
        package = Package('dif', self.distro, self.series, self.dest, self.ppa)
        ppa_source, dest_source = Mock(), Mock(component_name='universe')
        ppa_source.sourceFileUrls.return_value = [
            'ppa.tar', 'ppa.changes']
        dest_source.sourceFileUrls.return_value = [
            'dest.tar', 'dest.dsc', 'dest.changes']
        self.ppa.getPublishedSources.return_value = [ppa_source]
        self.dest.getPublishedSources.return_value = [dest_source]
        self.run_coro(package.diff())
        pre = DIFF.mock_calls[0][2]['env']
        self.assertEqual(pre['SOURCE'], 'dif')
        self.assertEqual(pre['SERIES'], 'xenial')
        self.assertEqual(pre['ACTION'], 'diff_prefetch')
        self.assertEqual(pre['DSC_URL'], '')
        dest = DIFF.mock_calls[1][2]['env']
        self.assertEqual(dest['SOURCE'], 'dif')
        self.assertEqual(dest['SERIES'], 'xenial')
        self.assertEqual(dest['ACTION'], 'diff_dest')
        self.assertEqual(dest['DSC_URL'], 'dest.dsc')
        self.assertEqual(dest['COMPONENT'], 'universe')
        self.assertSequenceEqual(DIFF.mock_calls, [
            call(join(Config.ROOT, 'scripts', 'vcs.sh'),
                 env=pre, stdout=PIPE, stderr=PIPE),
            call(join(Config.ROOT, 'scripts', 'vcs.sh'),
                 env=dest, stdout=PIPE, stderr=PIPE),
        ])
        self.assertSequenceEqual(self.ppa.getPublishedSources.mock_calls, [
            call(source_name='dif', distro_series=self.series,
                 exact_match=True, order_by_date=True),
        ])
        self.assertSequenceEqual(self.dest.getPublishedSources.mock_calls, [
            call(source_name='dif', distro_series=self.series,
                 exact_match=True, order_by_date=True),
        ])

    @patch('bileto.worker.vcs.create_subprocess_exec', POST_DIFF)
    def test_package_post_diff(self):
        """Invoke swift diff uploading."""
        ticket = Mock(request_id='5')
        Package.post_diff(ticket)
        env = POST_DIFF.mock_calls[0][2]['env']
        self.assertEqual(env['REQUESTID'], '5')
        self.assertEqual(env['ACTION'], 'swift_upload')
        self.assertSequenceEqual(POST_DIFF.mock_calls, [
            call(join(Config.ROOT, 'scripts', 'vcs.sh'),
                 env=env, stdout=PIPE, stderr=PIPE),
        ])
        self.assertEqual(ticket.mock_calls, [
            call.commit(),
        ])
        self.assertEqual(ticket.artifacts, 'swift/diff (9000 lines)')

    def test_package_post_abandon(self):
        """Set abandonment status."""
        ticket = Mock(request_id='5')
        Package.post_abandon(ticket)
        self.assertEqual(ticket.mock_calls, [
            call.set_status('Abandoned'),
        ])
