"""Bileto Tests

Define BiletoTestCase class, the superclass of all test classes, which
defines setUp and tearDown for all tests.
"""

from asyncio import coroutine, get_event_loop
from unittest import TestCase
from unittest.mock import Mock

from shutil import rmtree
from os import close, unlink
from tempfile import mkdtemp, mkstemp

from bileto.settings import Config
from bileto.app import app, db
from bileto.lplib import lp
from bileto.worker.log import log


def mock_coro(retval=None, retlist=None):
    """Create a coroutine for mocking other coroutines."""
    @coroutine
    def _mock_coro(*args, **kwargs):
        """This is the coroutine that gets used in place of others."""
        return retval or retlist.pop()
    return Mock(wraps=_mock_coro)


def fake_subproc(returncode, retlist):
    """Create a couple coroutines capable of mocking a subprocess."""
    return mock_coro(retval=Mock(
        returncode=returncode,
        communicate=mock_coro(retlist=retlist)))


class BiletoTestCase(TestCase):
    """Superclass of all test classes."""
    maxDiff = None

    def setUp(self):
        """Create temp database."""
        self.tempdir = mkdtemp()
        self.db_fd, filename = mkstemp()
        Config.HOME = self.tempdir
        app.config['DATABASE'] = filename
        app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///' + filename
        app.config['TESTING'] = True
        app.secret_key = 'test_secret'
        self.app = app.test_client()
        db.create_all()
        lp.__class__._real_instance = Mock(
            distributions=dict(ubuntu=Mock()),
            people=dict(us=Mock()),
            resource_type_link='https://api.launchpad.net/devel/#service-root',
        )
        self.distro = 'ubuntu'
        self.series = Mock()
        self.series.name = 'xenial'
        self.series.version = '16.04'
        self.dest = Mock()
        self.ppa = Mock()
        self.ppa.name = '12321'

    def tearDown(self):
        """Close & delete stuff."""
        close(self.db_fd)
        unlink(app.config['DATABASE'])
        rmtree(self.tempdir)

    def run_coro(self, coro):
        """Run a coroutine synchronously for testing."""
        return get_event_loop().run_until_complete(coro)


class WorkerTestCase(BiletoTestCase):
    """A few tweaks specific to bileto.worker tests."""

    def setUp(self):
        super().setUp()
        log.__class__.logdir.fget.cache_clear()
        log._dir = self.tempdir
        log._requestid = '1'
        log._action = 'build'
